#include <set>
#include <iostream>
#include <stdexcept>

#include <boost/foreach.hpp>

#include "configuration.hpp"
#include "fcf/net/socket.hpp"



namespace application {
namespace {

  std::string resolveString(boost::property_tree::ptree& a_ptree,
                            const char* a_field_name,
                            bool a_valid_empty,
                            bool a_empty_as_default,
                            bool a_valid_null,
                            std::string a_default
                            ){
    bool exists = a_ptree.find(a_field_name) != a_ptree.not_found();
    if (!exists && a_valid_null)
      return a_default;
    else if (!exists)
      throw std::runtime_error(std::string("The '") + a_field_name + "' config parameter is not set");
    std::string result = a_ptree.get<std::string>(a_field_name);
    if (result.empty() && a_valid_empty && a_empty_as_default)
      return a_default;
    else if (result.empty() && a_valid_empty)
      return result;
    else if (result.empty() && !a_valid_empty)
      throw std::runtime_error(std::string("The '") + a_field_name + "' config parameter is empty");
    return result;
  }

  int resolveInteger( boost::property_tree::ptree& a_ptree,
                      const char* a_field_name,
                      bool a_valid_null,
                      int  a_min,
                      int  a_max,
                      int a_default
                            ){
    bool exists = a_ptree.find(a_field_name) != a_ptree.not_found();
    if (!exists && a_valid_null)
      return a_default;
    else if (!exists)
      throw std::runtime_error(std::string("The '") + a_field_name + "' config parameter is not set");
    int result = 0;
    try {
      result = a_ptree.get<int>(a_field_name);
    } catch(const std::exception&){
      throw std::runtime_error(std::string("The '") + std::string(a_field_name) + "' parameter must be of the integer type");
    }
    if (a_min != -1 && result < a_min){
      throw std::runtime_error(std::string("The '") + std::string(a_field_name) + "' configuration parameter must be greater than " + std::to_string(a_min));
    }
    if (a_max != -1 && result > a_max){
      throw std::runtime_error(std::string("The '") + std::string(a_field_name) + "' configuration parameter must be less than " + std::to_string(a_max));
    }
    return result;
  }

  int resolveBoolean( boost::property_tree::ptree& a_ptree,
                      const char* a_field_name,
                      int a_default
                      ){
    bool exists = a_ptree.find(a_field_name) != a_ptree.not_found();
    if (!exists)
      return a_default;
    bool result = false;
    try {
      result = a_ptree.get<bool>(a_field_name);
    } catch(const std::exception&){
      throw std::runtime_error(std::string("The ") + std::string(a_field_name) + " parameter must be of the boolean type");
    }
    return result;
  }

  void fillPorts(std::vector<int>& a_vector, int a_count){
    int lastPort = a_vector[a_vector.size()-1];
    for(int i = (int)a_vector.size(); i < a_count; ++i){
      ++lastPort;
      a_vector.push_back(lastPort);
    }
  }


} // non namespace


  void handler_configuration_t::set_values(boost::property_tree::ptree& a_ptree, endpoint_configuration_t& a_owner) {
    this->innerControlPort       = a_owner.innerControlPort;
    this->user                    = resolveString(a_ptree, "user", true, true, true, a_owner.user);
    this->password                = resolveString(a_ptree, "password", true, true, true, a_owner.password);
    this->listenQueue             = resolveInteger(a_ptree, "listenQueue", true, 1, -1, a_owner.listenQueue);
    this->clientSocketTimeout     = resolveInteger(a_ptree, "clientSocketTimeout", true, 0, -1, a_owner.clientSocketTimeout);
    this->maxRequestTimeout       = resolveInteger(a_ptree, "maxRequestTimeout", true, 0, -1, a_owner.maxRequestTimeout);
    this->serverSocketTimeout     = resolveInteger(a_ptree, "serverSocketTimeout", true, 0, -1, a_owner.serverSocketTimeout);
    this->maxResponseTimeout      = resolveInteger(a_ptree, "maxResponseTimeout", true, 0, -1, a_owner.maxResponseTimeout);
    this->startTimeout            = resolveInteger(a_ptree, "startTimeout", true, 0, -1, a_owner.startTimeout);
    this->stopTimeout             = resolveInteger(a_ptree, "stopTimeout", true, 0, -1, a_owner.stopTimeout);
    this->keepAliveTimeout        = resolveInteger(a_ptree, "keepAliveTimeout", true, 0, -1, a_owner.keepAliveTimeout);

    this->name                    = resolveString(a_ptree, "name", false, false, false, "");
    if (this->name.length() > 16)
      throw std::runtime_error("The 'name' parameter is longer than 16 characters");
    this->host                    = resolveString(a_ptree, "host", true, true, true, this->host);
    this->path                    = resolveString(a_ptree, "path", true, true, true, this->path);
    this->processes               = resolveInteger(a_ptree, "processes", false, 1, 256, 1);
    this->processesLongRequests   = resolveInteger(a_ptree, "processesLongRequests", true, 0, 256, 0);
    if (!this->processes)
      throw std::runtime_error("The total number of processes of the handler is zero");

    this->script                  = resolveString(a_ptree, "script", false, false, false, "");
    if (a_ptree.find("scriptArgs") != a_ptree.not_found()){
      for(auto &p : a_ptree.get_child("scriptArgs")) {
        this->scriptArgs.push_back(p.second.data());
      }
    }
    this->maxMemoryLongRequest    = resolveInteger(a_ptree, "maxMemoryLongRequest", true, 0, -1, 150);
    this->maxMemory               = resolveInteger(a_ptree, "maxMemory", true, 0, -1, 150);

    bool data_ports_is_empty = true;
    if (a_ptree.find("dataPorts") != a_ptree.not_found()){
      for(auto &p : a_ptree.get_child("dataPorts")) {
        int port = 0;
        try {
          port = std::stoi(p.second.data());
          if (port != 0)
            data_ports_is_empty = false;
        } catch(...) {
        }
      }
    }
    if (data_ports_is_empty){
      this->dataPorts.clear();
      for (size_t i = 0; i < (this->processes + this->processesLongRequests) * 2; ++i)
        this->dataPorts.push_back(0);
    } else {
      this->dataPorts.clear();
      try {
        for(auto &p : a_ptree.get_child("dataPorts")){
          this->dataPorts.push_back(std::stoi(p.second.data()));
        }
      } catch(const std::exception&) {
        throw std::runtime_error("The dataPorts config parameter must be an array of integers");
      }
      fillPorts(this->dataPorts, (this->processes + this->processesLongRequests) * 2);
    }

    this->nodeOptions.clear();
    if (a_ptree.find("nodeOptions") != a_ptree.not_found()) {
      boost::property_tree::ptree ptNodeOptions = a_ptree.get_child("nodeOptions");
      try {
        BOOST_FOREACH(auto &ptNodeOptionsItem, ptNodeOptions) {
          this->nodeOptions.push_back(std::vector<std::string>());
          BOOST_FOREACH(auto &p, ptNodeOptionsItem.second) {
            this->nodeOptions[this->nodeOptions.size()-1].push_back(p.second.data());
          }
        }
      } catch(const std::exception&) {
        throw std::runtime_error("The nodeOptions config parameter must be an array of strings");
      }
    }
    for(size_t i = this->nodeOptions.size(); i < this->processes*2 + this->processesLongRequests*2; ++i)
      this->nodeOptions.push_back(std::vector<std::string>());
  }

  void endpoint_configuration_t::set_values(boost::property_tree::ptree& a_ptree, server_configuration_t& a_owner){
    this->innerControlPort       = a_owner.innerControlPort;
    this->user                    = resolveString(a_ptree, "user", true, true, true, a_owner.user);
    this->password                = resolveString(a_ptree, "password", true, true, true, a_owner.password);
    this->listenQueue             = resolveInteger(a_ptree, "listenQueue", true, 1, -1, a_owner.listenQueue);
    this->clientSocketTimeout     = resolveInteger(a_ptree, "clientSocketTimeout", true, 0, -1, a_owner.clientSocketTimeout);
    this->maxRequestTimeout       = resolveInteger(a_ptree, "maxRequestTimeout", true, 0, -1, a_owner.maxRequestTimeout);
    this->serverSocketTimeout     = resolveInteger(a_ptree, "serverSocketTimeout", true, 0, -1, a_owner.serverSocketTimeout);
    this->maxResponseTimeout      = resolveInteger(a_ptree, "maxResponseTimeout", true, 0, -1, a_owner.maxResponseTimeout);
    this->startTimeout            = resolveInteger(a_ptree, "startTimeout", true, 0, -1, a_owner.startTimeout);
    this->stopTimeout             = resolveInteger(a_ptree, "stopTimeout", true, 0, -1, a_owner.stopTimeout);
    this->transmissionBufferSize  = resolveInteger(a_ptree, "transmissionBufferSize", true, 0, -1, a_owner.transmissionBufferSize);
    this->keepAliveTimeout        = resolveInteger(a_ptree, "keepAliveTimeout", true, 0, -1, a_owner.keepAliveTimeout);
    this->bind4                   = resolveString(a_ptree, "bind4", true, false, true, this->bind4);
    this->bind6                   = resolveString(a_ptree, "bind6", true, true, true, this->bind6);
    this->port                    = resolveInteger(a_ptree, "port", false, 1, 65535, this->port);

    if (a_ptree.find("handlers") == a_ptree.not_found())
      throw std::runtime_error("The 'handlers' config parameter is not set");
    boost::property_tree::ptree handlers = a_ptree.get_child("handlers");
    if (handlers.size() == 0)
      throw std::runtime_error("'handlers' config parameter must contain an array of objects");
    BOOST_FOREACH(auto &h, handlers){
      handler_configuration_t hc;
      hc.set_values(h.second, *this);
      this->handlers.push_back(hc);
    }
  }


  void server_configuration_t::set_values(boost::property_tree::ptree& a_ptree, const global_configuration_t& a_gc){
    this->logFile                 = resolveString(a_ptree, "logFile", true, true, true, std::string(""));
    std::string logLevel          = resolveString(a_ptree, "logLevel", true, true, true, std::string("log"));
    if      (logLevel == "err") this->logLevel = fcf::LOG_LEVEL_ERR;
    else if (logLevel == "wrn") this->logLevel = fcf::LOG_LEVEL_WRN;
    else if (logLevel == "trc") this->logLevel = fcf::LOG_LEVEL_TRC;
    else                        this->logLevel = fcf::LOG_LEVEL_LOG;
    this->logLifeTime             = resolveInteger(a_ptree, "logLifeTime", true, 0, -1, this->logLifeTime);
    this->innerControlPort       = resolveInteger(a_ptree, "innerControlPort", true, 0, 65535, this->innerControlPort);
    if (!this->innerControlPort)
      this->innerControlPort     = fcf::net::get_free_tcp_port("Server control port (innerControlPort parameter)", a_gc.free_ports);
    this->user                    = resolveString(a_ptree, "user", true, true, true, this->user);
    this->password                = resolveString(a_ptree, "password", true, true, true, this->password);
    this->listenQueue             = resolveInteger(a_ptree, "listenQueue", true, 1, -1, this->listenQueue);
    this->clientSocketTimeout     = resolveInteger(a_ptree, "clientSocketTimeout", true, 0, -1, this->clientSocketTimeout);
    this->maxRequestTimeout       = resolveInteger(a_ptree, "maxRequestTimeout", true, 0, -1, this->maxRequestTimeout);
    this->serverSocketTimeout     = resolveInteger(a_ptree, "serverSocketTimeout", true, 0, -1, this->serverSocketTimeout);
    this->maxResponseTimeout      = resolveInteger(a_ptree, "maxResponseTimeout", true, 0, -1, this->maxResponseTimeout);
    this->startTimeout            = resolveInteger(a_ptree, "startTimeout", true, 0, -1, this->startTimeout);
    this->stopTimeout             = resolveInteger(a_ptree, "stopTimeout", true, 0, -1, this->stopTimeout);
    this->transmissionBufferSize  = resolveInteger(a_ptree, "transmissionBufferSize", true, 0, -1, this->transmissionBufferSize);
    this->keepAliveTimeout        = resolveInteger(a_ptree, "keepAliveTimeout", true, 0, -1, this->keepAliveTimeout);

    if (a_ptree.find("endpoints") == a_ptree.not_found())
      throw std::runtime_error("The 'endpoints' config parameter is not set");
    boost::property_tree::ptree ptendpoints = a_ptree.get_child("endpoints");
    if (ptendpoints.size() == 0)
      throw std::runtime_error("The 'endpoints' config parameter must contain an array of objects");
    for(auto& ep : ptendpoints){
      endpoint_configuration_t epc;
      epc.set_values(ep.second, *this);
      endpoints.push_back(epc);
    }
    std::set<std::string> names;
    for(endpoint_configuration_t& ep : endpoints) {
      for(handler_configuration_t& handler : ep.handlers) {
        if (names.find(handler.name) != names.end())
          throw std::runtime_error("The handler name has a duplicate in the config file");
        names.insert(handler.name);
      }
    }

    if (a_ptree.find("servers") == a_ptree.not_found())
      throw std::runtime_error("The 'servers' config parameter is not set");
    boost::property_tree::ptree tpserver_connections = a_ptree.get_child("servers");
    for(auto& ep : tpserver_connections){
      server_connection_t sc;
      sc.name    = resolveString(ep.second, "name", false, false, false, "");
      if (sc.name.length() > 32)
        throw std::runtime_error("The 'name' parameter in the 'servers' option is longer than 32 characters");
      sc.address = resolveString(ep.second, "address", true, true, true, "");
      sc.port = resolveInteger(ep.second, "port", false, 1, 65535, 0);
      this->servers.push_back(sc);
    }

    this->serverName  = resolveString(a_ptree, "serverName", false, false, false, "");
    if (this->serverName.length() > 32)
      throw std::runtime_error("The 'serverName' parameter is longer than 32 characters");
    this->mainServer  = resolveString(a_ptree, "mainServer", false, false, false, "");
    if (this->mainServer.length() > 32)
      throw std::runtime_error("The 'mainServer' parameter is longer than 32 characters");

  }


}
