#ifdef WIN32
#pragma warning(disable:4996)
#endif
#include <iostream>
#include <sstream>
#include <fstream>
#include <filesystem>
#include <thread>
#include <codecvt>
#include <climits>
#ifdef __APPLE__
#include <libproc.h>
#endif

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/thread.hpp>
#include <boost/process/search_path.hpp>

#include <fcf/except.hpp>
#include <fcf/system.hpp>
#include <fcf/log.hpp>
#include <fcf/filesystem.hpp>
#include <fcf/net/protocol/http.hpp>

#include "tools.hpp"
#include "configuration.hpp"
#include "endpoint.hpp"

#ifdef _WIN32
#include <windows.h>
#else
#include <limits.h>
#include <unistd.h>
#endif


namespace application {

#define MAX_PACK   64*1024
#define STACK_SIZE MAX_PACK*4



namespace {

  std::filesystem::path get_temporary_path(){
    std::filesystem::path path = std::filesystem::temp_directory_path();
    std::stringstream ss;
    ss << "fcf-framework-transfer-data-" << boost::uuids::random_generator()();
    path /= ss.str().c_str();
    return path;
  }

  class file_buffer_t {
  public:
    file_buffer_t()
      : _path(get_temporary_path()), _offset(0), _stream(_path, std::fstream::out | std::fstream::in | std::fstream::app | std::fstream::binary) {
      _stream.exceptions(std::ios::badbit | std::ios::failbit);
    }

    ~file_buffer_t(){
      try { _stream.close(); } catch (std::exception e) {}
      try {
        std::filesystem::remove(_path);
      } catch (std::exception e) {
        fcf::err << "ERROR REMOVE FILE TEMP FILE " << _path.string();
      }
    }

    void write(const char* a_datav, size_t a_datac) {
      _stream.write(a_datav, a_datac);
    }

    void read(char* a_buffer, size_t a_buffer_size) {
      _stream.seekg(_offset);
      _stream.read(a_buffer, a_buffer_size);
      _offset += a_buffer_size;
    }

  private:
    std::filesystem::path _path;
    unsigned long long    _offset;
    std::fstream          _stream;
  };

}

  endpoint_t::endpoint_t(const server_configuration_t& a_server_config, const global_configuration_t& a_global_config, size_t a_endpoint_index)
    : _server_name(a_server_config.serverName),
      _main_server_name(a_server_config.mainServer),
      _config(a_server_config.endpoints[a_endpoint_index]),
      _global_config(a_global_config),
      _stop(false) {
    if (_config.bind4 != "none")
      _socket4.reset(fcf::net::create_socket(_config.bind4.c_str(), _config.port));
    if (_config.bind6 != "none")
      _socket6.reset(fcf::net::create_socket(_config.bind6.c_str(), _config.port));
  }

  endpoint_t::~endpoint_t(){
    try {
      stop();
    } catch (...){
    }
  }

  void endpoint_t::run() {
    std::unique_lock<std::mutex> lock(_launch_mutex);
    fcf::log << "Running processes...";
    for(size_t hi = 0; hi < _config.handlers.size(); ++hi) {
      const handler_configuration_t& hc(_config.handlers[hi]);

      subprocesses_sptr_t spsp(new subprocesses_t(_global_config, hc.maxMemory, hc.maxMemoryLongRequest, hc.name, _server_name));
      _subprocesses.insert({hc.name, spsp});

      size_t procindex = 0;
      for(int long_req = 0; long_req <= 1; ++long_req) {
        int procc = long_req ? hc.processesLongRequests: hc.processes;
        for(size_t i = 0; i < procc; ++i) {
          std::vector<std::string> argv1;
          argv1.push_back(hc.script);
          if (_main_server_name != _server_name)
            argv1.push_back("--slave");
          argv1.push_back("--server-name");
          argv1.push_back(_server_name);
          argv1.push_back("--main-server-name");
          argv1.push_back(_main_server_name);
          argv1.push_back("--handler-name");
          argv1.push_back(hc.name);
          argv1.push_back("--process-index");
          argv1.push_back(std::to_string(procindex*2));
          argv1.push_back("--inner-control-port");
          argv1.push_back(std::to_string(hc.innerControlPort));
          argv1.push_back("--port");
          argv1.push_back(std::to_string(hc.dataPorts[procindex*2]));
          argv1.push_back("--host");
          argv1.push_back("127.0.0.1");
          argv1.push_back("--keep-alive-timeout");
          argv1.push_back(std::to_string(hc.keepAliveTimeout ? hc.keepAliveTimeout + 5: 0));

          if (!!procindex) {
            argv1.push_back("--disable-sys");
            argv1.push_back("--disable-cron");
          } else {
            argv1.push_back("--leading");
          }
          if (!long_req)
            argv1.push_back("--leak-protected");
          for(auto& arg : hc.scriptArgs){
            argv1.push_back(arg);
          }


          std::vector<std::string> argv2;
          argv2.push_back(hc.script);
          if (_main_server_name != _server_name)
            argv2.push_back("--slave");
          argv2.push_back("--server-name");
          argv2.push_back(_server_name);
          argv2.push_back("--main-server-name");
          argv2.push_back(_main_server_name);
          argv2.push_back("--handler-name");
          argv2.push_back(hc.name);
          argv2.push_back("--process-index");
          argv2.push_back(std::to_string(procindex*2+1));
          argv2.push_back("--inner-control-port");
          argv2.push_back(std::to_string(hc.innerControlPort));
          argv2.push_back("--port");
          argv2.push_back(std::to_string(hc.dataPorts[procindex*2+1]));
          argv2.push_back("--host");
          argv2.push_back("127.0.0.1");
          argv2.push_back("--keep-alive-timeout");
          argv2.push_back(std::to_string(hc.keepAliveTimeout ? hc.keepAliveTimeout + 10: 0));
          if (!!procindex) {
            argv2.push_back("--disable-sys");
            argv2.push_back("--disable-cron");
          } else {
            argv2.push_back("--leading");
          }
          if (!long_req)
            argv2.push_back("--leak-protected");
          for(auto& arg : hc.scriptArgs){
            argv2.push_back(arg);
          }

          std::vector<std::pair<std::string, std::string> > envv;
          std::wstring_convert<std::codecvt_utf8<wchar_t>> conv;

          #ifdef WIN32
          std::string exec_path_str = "node";
          std::string nodePath = "./node_modules;" + fcf::get_exe_directory().string() + ";" + fcf::get_exe_directory().string() + "\\" + "node_modules";
          envv.push_back(std::pair<std::string, std::string>("NODE_PATH", nodePath));
          #else
          std::filesystem::path exec_path = fcf::get_exe_directory() / "fcfnode";
          std::string           exec_path_str = conv.to_bytes(exec_path.wstring());
          #endif
          envv.push_back(std::pair<std::string, std::string>("PATH", std::getenv("PATH")));

          auto get_node_options = [&hc, long_req](const std::vector<std::string>& a_src) {
            std::vector<std::string> result;
            bool ignore_max_old_space_size = false;
            for(const std::string& arg : a_src){
              if (arg.find("--max-old-space-size") != std::string::npos) {
                ignore_max_old_space_size = true;
                break;
              }
            }

            if (!ignore_max_old_space_size) {
              size_t max_memory = !long_req ? hc.maxMemory : hc.maxMemoryLongRequest;
              if (max_memory)
                result.push_back(std::string("--max-old-space-size=") + std::to_string(max_memory < 100 ? (size_t)max_memory + 100 : (size_t)(1.5 * max_memory + 50)));
            }

            for(const std::string& arg : a_src)
              result.push_back(arg);

            return result;
          };
          std::vector<std::string> node_options1(get_node_options(hc.nodeOptions[procindex*2]));
          std::vector<std::string> node_options2(get_node_options(hc.nodeOptions[procindex*2 + 1]));

          _subprocesses[hc.name]->append(exec_path_str.c_str(),
                                         argv1,
                                         argv2,
                                         envv,
                                         node_options1,
                                         node_options2,
                                         hc.user.c_str(),
                                         hc.password.c_str(),
                                         hc.startTimeout,
                                         hc.stopTimeout,
                                         !long_req,
                                         !procindex);
          ++procindex;
        }
      }
      _subprocesses[hc.name]->run();
    }

    if (_socket4.get()){
      fcf::log << "Listening started " << _config.bind4 << " with port " << _config.port;
      _socket4->set_reuse(true);
      _socket4->bind();
      _socket4->listen(_config.listenQueue);
      _endpoint_thread4 = std::thread(&endpoint_t::_endpoint_handler, std::ref(*this), _socket4.get());
    }

    if (_socket6.get()){
      fcf::log << "Listening started " << _config.bind6 << " with port " << _config.port;
      _socket6->set_reuse(true);
      _socket6->bind();
      _socket6->listen(_config.listenQueue);
      _endpoint_thread6 = std::thread(&endpoint_t::_endpoint_handler, std::ref(*this), _socket6.get());
    }

  }

  void endpoint_t::stop() {
    std::unique_lock<std::mutex> lock(_launch_mutex);
    if (_stop)
      return;

    _stop = true;

    if (_socket4.get())
      _socket4->close();
    if (_socket6.get())
      _socket6->close();

    if (_endpoint_thread4.joinable())
      _endpoint_thread4.join();
    if (_endpoint_thread6.joinable())
      _endpoint_thread6.join();

    int print_count = 0;
    for(size_t hi = 0; hi < _config.handlers.size(); ++hi) {
      if (_subprocesses[_config.handlers[hi].name].get()){
        _subprocesses[_config.handlers[hi].name]->stop();
      }
    }
  }

  void endpoint_t::restart_handler(std::string a_handler){
    std::map<std::string, subprocesses_sptr_t>::iterator it = _subprocesses.find(a_handler);
    if (it != _subprocesses.end())
      it->second->restart();
  }

  void endpoint_t::restart_handlers(){
    for(std::pair<const std::string, subprocesses_sptr_t>& item : _subprocesses){
      item.second->restart();
    }
  }

  void endpoint_t::get_processes_for_message(std::vector<process_identification_t>& a_dst,
                                             bool a_only_active_processes,
                                             bool a_active_flag,
                                             int a_destenition_pair,
                                             std::string a_handler,
                                             std::string a_ignore_handler,
                                             int a_ignore_process){
    for (auto& [key, psubprocess] : _subprocesses) {
      psubprocess->get_processes_for_message(a_dst,
                                             a_only_active_processes,
                                             a_active_flag,
                                             a_destenition_pair,
                                             a_handler,
                                             a_ignore_handler,
                                             a_ignore_process);
    }
  }

  void endpoint_t::release_processes_for_message(const std::vector<process_identification_t>& a_procs, bool a_active){
    for (auto& [key, psubprocess] : _subprocesses)
      psubprocess->release_processes_for_message(a_procs, a_active);
  }

  void endpoint_t::_endpoint_handler(fcf::net::socket_t* a_socket){
    while(!_stop){
      fcf::net::socket_t* socket = 0;
      try {
        fcf::net::socket_sptr_t socket = fcf::net::socket_sptr_t(a_socket->accept());
        boost::thread::attributes attrs;
        attrs.set_stack_size(STACK_SIZE);
        boost::thread(attrs, boost::bind(&endpoint_t::_handler, this, socket)).detach();
      } catch(std::exception& e){
        if (!_stop)
          fcf::err << "Accept error: " << e.what();
      } catch (...) {
        if (!_stop)
          fcf::err << "Unknown Exception in accept";
      }

    }
  }

  void endpoint_t::_handler(fcf::net::socket_sptr_t a_socket){
    int                                     offset = 0;
    char                                    big_data_buffer[MAX_PACK];
    char                                    first_buffer[MAX_PACK*2];
    int                                     first_buffer_offset = 0;
    std::vector<char>                       residual_buffer;
    unsigned long long                      total_size = 0;
    unsigned long long                      rec_total_size = 0;
    fcf::net::socket_sptr_t                 process_socket;
    fcf::net::protocol::http_header_info_t  request_header_info;
    fcf::net::protocol::http_header_info_t  response_header_info;
    bool                                    short_mode = false;
    int                                     handler_index = -1;
    int                                     process_index = -1;
    unsigned long long                      full_pack;
    bool                                    first_transaction = true;
    bool                                    completed = false;
    bool                                    is_big_data = false;
    subprocesses_t::subprocesses_state_t*   sps = 0;
    std::shared_ptr<file_buffer_t>          file_buffer;
    int                                     client_socket_timeout;
    int                                     max_request_timeout;
    int                                     server_socket_timeout;
    int                                     max_response_timeout;

    boost::uuids::uuid                      debug_client_uuid = boost::uuids::random_generator()();
    bool                                    request_timeout_error = false;

    try {
      while(true) {
        std::chrono::time_point start_request = std::chrono::system_clock::now();

        client_socket_timeout = _config.clientSocketTimeout;
        max_request_timeout   = _config.maxRequestTimeout;
        server_socket_timeout = _config.serverSocketTimeout;
        max_response_timeout  = _config.maxResponseTimeout;

        if (first_transaction)
          a_socket->set_timeout(client_socket_timeout);

        bool first_request = true;
        // read header
        bool break_loop = false;
        while(true) {

          completed = false;

          if (!first_transaction && first_request)
            a_socket->set_timeout(_config.keepAliveTimeout);

          int recc = 0;
          recc = a_socket->recv(&first_buffer[first_buffer_offset], MAX_PACK);
          if (!recc) {
            if (!first_transaction && first_request && std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() >= (_config.keepAliveTimeout)){
              break_loop = true;
              request_timeout_error = true;
              break;
            } else if (first_request && std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() >= (_config.keepAliveTimeout)){
              throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());
            } else if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout) {
              throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());
            }
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
            continue;
          }

          if (!first_transaction && first_request)
            a_socket->set_timeout(client_socket_timeout);

          first_request = false;

          if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
            throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());

          first_buffer_offset += recc;
          total_size = first_buffer_offset;
          rec_total_size = first_buffer_offset;
          if (!fcf::net::protocol::parse_http_header_request(first_buffer, first_buffer_offset, &request_header_info)){
            if (first_buffer_offset > MAX_PACK)
              throw std::runtime_error("Request header size exceeds maximum packet size");
            continue;
          }
          break;
        }

        if (break_loop)
          break;

        fcf::trc << "Request: " << a_socket->get_address() << ": PORT:" << a_socket->get_port() << "; Path: " << request_header_info.path;

        first_transaction = false;

        // find handler configuration
        // width  0 - host and path
        //        1 - host and path (partially) (the 'path_cmp_weigth' variable is taken into account)
        //        2 - host and path (none)
        //        3 - path
        //        4 - path (partially) (the 'path_cmp_weigth' variable is taken into account)
        //        5 - all
        int path_cmp_weigth = 0;
        int width = INT_MAX;
        int handlerc = (int)_config.handlers.size();
        int new_handler_index = -1;
        for (int hi = 0; hi < handlerc; ++hi) {
          const handler_configuration_t& hconf(_config.handlers[hi]);
          bool cur_host_empty = hconf.host.empty();
          bool cur_host_iseq = hconf.host == request_header_info.host;

          bool cur_path_empty = hconf.path.empty();
          bool cur_path_iseq = false;
          int  cur_path_cmp_width = 0;
          bool cur_path_match = !cur_path_empty ? cmp_path_template(hconf.path.c_str(), request_header_info.host.c_str(), &cur_path_iseq, &cur_path_cmp_width)
                                                : true;

          if (!cur_host_empty && cur_host_iseq &&
              !cur_path_empty && cur_path_iseq) {
            new_handler_index = hi;
            width = 0;
            path_cmp_weigth = 0;
            break;
          } else if (width >= 1 &&
                    !cur_host_empty && cur_host_iseq &&
                    !cur_path_empty && cur_path_match &&
                    (width != 1 || path_cmp_weigth > cur_path_cmp_width)) {
            new_handler_index = hi;
            width = 1;
            path_cmp_weigth = cur_path_cmp_width;
          } else if (width > 2 &&
                    !cur_host_empty && cur_host_iseq &&
                    cur_path_empty) {
            new_handler_index = hi;
            width = 2;
            path_cmp_weigth = 0;
          } else if (width > 3 &&
                    cur_host_empty &&
                    !cur_path_empty && cur_path_iseq){
            new_handler_index = hi;
            width = 3;
            path_cmp_weigth = 0;
          } else if (width >= 4 &&
                    cur_host_empty &&
                    (width != 4 || path_cmp_weigth > cur_path_cmp_width)){
            new_handler_index = hi;
            width = 4;
            path_cmp_weigth = cur_path_cmp_width;
          } else if (width > 5 &&
                    cur_host_empty &&
                    cur_path_empty){
            new_handler_index = hi;
            width = 5;
            path_cmp_weigth = 0;
          }
        }

        if (new_handler_index == -1)
          throw std::runtime_error(std::string(std::string("Missing request handler for host '") + request_header_info.host + "' with path '" + request_header_info.path + "'").c_str());

        // reading the residual message
        full_pack   = request_header_info.content_length + request_header_info.header_end + 4;
        is_big_data = full_pack >= _config.transmissionBufferSize;
        if (!is_big_data) {
          if (first_buffer_offset < full_pack)
            residual_buffer.resize(full_pack - first_buffer_offset);
          while (rec_total_size < full_pack) {
            size_t size = a_socket->recv(&residual_buffer[rec_total_size - first_buffer_offset], full_pack - rec_total_size < MAX_PACK ? full_pack - rec_total_size : MAX_PACK);
            if (!size)
              std::this_thread::sleep_for(std::chrono::milliseconds(10));
            rec_total_size += size;
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }
        } else if (rec_total_size < full_pack) {
          file_buffer.reset(new file_buffer_t());
          while (rec_total_size < full_pack) {
            size_t size = a_socket->recv(&big_data_buffer[0], full_pack - rec_total_size < MAX_PACK ? full_pack - rec_total_size : MAX_PACK);
            if (!size)
              std::this_thread::sleep_for(std::chrono::milliseconds(10));
            rec_total_size += size;
            file_buffer->write(&big_data_buffer[0], size);
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }
        }

        // reconnect to process
        const handler_configuration_t& handler_config(_config.handlers[new_handler_index]);

        if (new_handler_index != handler_index || !_subprocesses[_config.handlers[handler_index].name]->is_active(process_index)) {
          int old_handler_index = handler_index;
          handler_index = new_handler_index;

          if (process_index != -1){
            _subprocesses[_config.handlers[old_handler_index].name]->release_process(process_index);
            process_index = -1;
          }


          process_index = _subprocesses[_config.handlers[handler_index].name]->get_process();

          sps = &(_subprocesses[_config.handlers[handler_index].name]->get_process_info(process_index));
          if (process_socket)
            process_socket->close();
          process_socket.reset(fcf::net::create_socket("127.0.0.1", sps->real_port));
          process_socket->set_timeout(server_socket_timeout);
          process_socket->connect();
        }

        // send header with data
        offset = 0;
        {
          while(offset < first_buffer_offset){
            size_t size = process_socket->send(&first_buffer[offset], first_buffer_offset - offset);
            offset += size;
            if (!size)
              std::this_thread::sleep_for(std::chrono::milliseconds(10));
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }
        }

        // send residual message to process
        offset = 0;
        if (!is_big_data) {
          //residual_buffer
          while((offset + first_buffer_offset) < full_pack) {
            size_t size = process_socket->send(&residual_buffer[offset], (full_pack - first_buffer_offset) - offset);
            offset += size;
            if (!size)
              std::this_thread::sleep_for(std::chrono::milliseconds(10));
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }
          residual_buffer.clear();
        } else if (first_buffer_offset < full_pack) {
          while((offset + first_buffer_offset) < full_pack) {
            size_t pack_size = (full_pack - first_buffer_offset) - offset < MAX_PACK ? (full_pack - first_buffer_offset) - offset : MAX_PACK;
            file_buffer->read(&big_data_buffer[0], pack_size);
            size_t part_size = 0;
            while(part_size < pack_size) {
              size_t size = process_socket->send(&big_data_buffer[part_size], pack_size - part_size);
              part_size += size;
              if (!size)
                std::this_thread::sleep_for(std::chrono::milliseconds(10));
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
                throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
            }
            offset += pack_size;
          }
        }

        // get response
        std::chrono::time_point start_response = std::chrono::system_clock::now();
        offset = 0;
        bool first = true;
        process_socket->set_timeout(max_response_timeout);
        while(true) {
          int recc = process_socket->recv(&first_buffer[offset], MAX_PACK);
          if (!recc)
            throw std::runtime_error("Failed to get data from server");

          if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
            throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());

          if (first)
            process_socket->set_timeout(server_socket_timeout);

          first = false;
          offset += recc;
          if (!fcf::net::protocol::parse_http_header_response(first_buffer, offset, &response_header_info)){
            if (offset > MAX_PACK)
              throw std::runtime_error("Response header size exceeds maximum packet size");
            continue;
          }
          total_size = offset;
          break;
        }

        // send header with data to client
        full_pack = response_header_info.content_length + response_header_info.header_end + 4;
        offset = 0;
        while (offset < total_size) {
          if (!sps->active) {
            _buffered_forwarding(&first_buffer[offset],
                                 total_size - offset,
                                 full_pack - offset,
                                 a_socket,
                                 process_socket,
                                 handler_index,
                                 process_index,
                                 start_response,
                                 request_header_info,
                                 max_response_timeout,
                                 big_data_buffer);
            completed = true;
            break;
          }
          size_t size = a_socket->send(&first_buffer[offset], total_size - offset > MAX_PACK ? MAX_PACK : (unsigned int)(total_size - offset));
          offset += size;
          if (!size)
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
          if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
            throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
        }

        // resend data to client
        if (!completed) {
          while(total_size < full_pack) {
            if (!sps->active) {
              _buffered_forwarding(&first_buffer[0],
                                   0,
                                   full_pack - total_size,
                                   a_socket,
                                   process_socket,
                                   handler_index,
                                   process_index,
                                   start_response,
                                   request_header_info,
                                   max_response_timeout,
                                   big_data_buffer);
              completed = true;
              break;
            }

            offset = process_socket->recv(&first_buffer[0], MAX_PACK);
            if (!offset)
              std::this_thread::sleep_for(std::chrono::milliseconds(10));
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
              throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());

            if (!sps->active) {
              _buffered_forwarding(&first_buffer[0],
                                   offset,
                                   full_pack - total_size,
                                   a_socket,
                                   process_socket,
                                   handler_index,
                                   process_index,
                                   start_response,
                                   request_header_info,
                                   max_response_timeout,
                                   big_data_buffer);
              completed = true;
              break;
            }

            int sendc = 0;
            {
              while(sendc < offset){
                size_t size = a_socket->send(&first_buffer[sendc], offset - sendc);
                sendc += size;
                if (!size)
                  std::this_thread::sleep_for(std::chrono::milliseconds(10));
                if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
                  throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
              }
            }

            total_size += offset;
          }
        }

        if (response_header_info.connection != "keep-alive"){
          break;
        }

        if (_stop)
          break;

        if (!_config.keepAliveTimeout)
          break;

        offset              = 0;
        first_buffer_offset = 0;
        total_size          = 0;
        completed           = false;
      }

    } catch(const fcf::except::max_request_timeout& e) {
      request_timeout_error = true;
      fcf::dbg << "IP:" << a_socket->get_address() << " PORT:" << a_socket->get_port() << " " << e.what();
    } catch(const std::exception& e) {
      fcf::dbg << "IP:" << a_socket->get_address() << " PORT:" << a_socket->get_port() << " " << e.what();
    } catch (...) {
      fcf::dbg << "Unknown Exception in handler";
    }

    if (request_timeout_error){
      const char* message = "HTTP 408 Request Timeout\r\n"
                            "Connection: close\r\n"
                            "\r\n\r\n";
      try {
        a_socket->set_timeout(1);
        a_socket->send(message, std::strlen(message));
      } catch(...){
      }
    }

    if (process_index != -1) {
      _subprocesses[_config.handlers[handler_index].name]->release_process(process_index);
      process_index = -1;
    }
  }

  void endpoint_t::_buffered_forwarding(const char*                             a_unsent_data_buffer,
                                        size_t                                  a_unsent_data_buffer_size,
                                        unsigned long long                      a_total_unset_data_size,
                                        fcf::net::socket_sptr_t                 a_client_socket,
                                        fcf::net::socket_sptr_t                 a_process_socket,
                                        int                                     a_handler_index,
                                        int&                                    a_process_index,
                                        std::chrono::system_clock::time_point   a_start_response,
                                        fcf::net::protocol::http_header_info_t& a_request_header_info,
                                        int                                     a_max_response_timeout,
                                        char*                                   a_big_data_buffer) {
    std::unique_ptr< std::vector<char> > mem_buffer;
    std::unique_ptr< file_buffer_t > file_buffer;
    unsigned long long buffer_size = a_total_unset_data_size - a_unsent_data_buffer_size;
    bool use_mem_buffer = buffer_size <= _config.transmissionBufferSize;
    if (use_mem_buffer){
      mem_buffer.reset(new std::vector<char>(buffer_size));
    } else {
      file_buffer.reset(new file_buffer_t());
    }

    unsigned long long offset = 0;
    while (offset < (a_total_unset_data_size - a_unsent_data_buffer_size)) {
      unsigned long long residual = (a_total_unset_data_size - a_unsent_data_buffer_size) - offset;
      size_t size = a_process_socket->recv(use_mem_buffer ? &(*mem_buffer)[offset] : a_big_data_buffer, residual < MAX_PACK ? residual : MAX_PACK);
      if (!size)
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
      if (!use_mem_buffer)
        file_buffer->write(a_big_data_buffer, size);
      offset += size;
      if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - a_start_response).count() > a_max_response_timeout)
        throw fcf::except::max_response_timeout_fi(a_client_socket->get_address(), a_client_socket->get_port(), a_request_header_info.host.c_str(), a_request_header_info.path.c_str());
    }

    if (a_process_index != -1 && a_handler_index != -1){
      _subprocesses[_config.handlers[a_handler_index].name]->release_process(a_process_index);
      a_process_index = -1;
    }

    offset = 0;
    while (offset < a_unsent_data_buffer_size) {
      size_t size = a_client_socket->send(&a_unsent_data_buffer[offset], a_unsent_data_buffer_size - offset < MAX_PACK ? a_unsent_data_buffer_size - offset : MAX_PACK);
      if (!size)
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
      offset += size;
      if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - a_start_response).count() > a_max_response_timeout)
        throw fcf::except::max_response_timeout_fi(a_client_socket->get_address(), a_client_socket->get_port(), a_request_header_info.host.c_str(), a_request_header_info.path.c_str());
    }

    while (offset < a_total_unset_data_size) {
      size_t size;
      size_t pack_size = a_total_unset_data_size - offset < MAX_PACK ? a_total_unset_data_size - offset : MAX_PACK;
      if (use_mem_buffer){
        size = a_client_socket->send(&(*mem_buffer)[offset - a_unsent_data_buffer_size], pack_size);
      } else {
        file_buffer->read(a_big_data_buffer, pack_size);
        size = a_client_socket->send(a_big_data_buffer, pack_size);
      }
      if (!size)
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
      offset += size;
      if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - a_start_response).count() > a_max_response_timeout)
        throw fcf::except::max_response_timeout_fi(a_client_socket->get_address(), a_client_socket->get_port(), a_request_header_info.host.c_str(), a_request_header_info.path.c_str());
    }
  }


}
