#include <sstream>
#include <time.h>
#include <stdlib.h>
#include <chrono>
#include <boost/algorithm/hex.hpp>
#include <boost/iostreams/device/array.hpp>
#include <boost/iostreams/stream.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include "fcf/log.hpp"
#include "fcf/sleep.hpp"
#include "fcf/net/net.hpp"
#include "server.hpp"
#include "event_server.hpp"

namespace application {

  void event_server_t::message_t::build_responce(message_t& a_message) {
    if (a_message.header.error) {
      a_message.message.resize(
        event_server_t::message_header_t::raw_header_size + a_message.error.size());
      memcpy(&a_message.message[event_server_t::message_header_t::raw_header_size], &a_message.error[0], a_message.error.size());
    } else {
      if (!a_message.header.sync_modify) {
        a_message.message.resize(event_server_t::message_header_t::raw_header_size);
        if (a_message.result.size()) {
          a_message.header.result = true;
          a_message.message.insert(a_message.message.end(), a_message.result.begin(), a_message.result.end());
        }
      }
    }

    a_message.header.body_size         = a_message.message.size() - message_header_t::raw_header_size;
    a_message.header.response          = true;
    a_message.header.outgoing_message  = false;

    if (a_message.message.size() < event_server_t::message_header_t::raw_header_size)
      a_message.message.resize(event_server_t::message_header_t::raw_header_size);

    message_header_t::to_raw(&a_message.message[0], a_message.header);
  }

  void event_server_t::message_header_t::from_raw(message_header_t& a_header, const char a_raw[event_server_t::message_header_t::raw_header_size]){
    const char* p = (const char*)&a_raw[0];
    a_header.body_size = fcf::net::le_to_cpu(*(unsigned int*)(p));
    p+=4;

    char bits = *(char*)(p);
    a_header.response                 = !!(bits & 0x01);
    a_header.destination_inner        = !!(bits & 0x02);
    a_header.sync_modify              = !!(bits & 0x04);
    a_header.error                    = !!(bits & 0x08);
    a_header.result                   = !!(bits & 0x10);
    a_header.enable_connection_error  = !!(bits & 0x20);
    a_header.active                   = !!(bits & 0x40);
    a_header.noactive                 = !!(bits & 0x80);
    p+=1;

    bits = *(char*)(p);
    a_header.ignore_self_noactive_calls  = !!(bits & 0x01);
    a_header.directed_call               = !!(bits & 0x02);
    a_header.from_active_call            = !!(bits & 0x04);
    a_header.lock_type                   = (lock_type_t)fcf::net::le_to_cpu((bits & 0b00111000 ) >> 3);

    p+=1;

    a_header.id.assign(p, 18);
    p+=18;

    a_header.destenition_server = "";
    for(size_t i = 0; i < 32 && !!p[i]; ++i)
      a_header.destenition_server += p[i];
    p+=32;

    a_header.destination_handler = "";
    for(size_t i = 0; i < 16 && !!p[i]; ++i)
      a_header.destination_handler += p[i];
    p+=16;

    a_header.destination_pair     = (int)fcf::net::le_to_cpu((*(short*)p));
    p+=2;
    a_header.lock_counter         = fcf::net::le_to_cpu((*(unsigned short*)p));

    a_header.outgoing_message     = false;
  }

  void event_server_t::message_header_t::to_raw(char a_raw[event_server_t::message_header_t::raw_header_size], const message_header_t& a_header){
    char* p = (char*)&a_raw[0];
    *(unsigned int*)p = fcf::net::le_to_cpu(a_header.body_size);
    p+=4;

    *p = (a_header.response                 ? 0x01 : 0) |
         (a_header.destination_inner        ? 0x02 : 0) |
         (a_header.sync_modify              ? 0x04 : 0) |
         (a_header.error                    ? 0x08 : 0) |
         (a_header.result                   ? 0x10 : 0) |
         (a_header.enable_connection_error  ? 0x20 : 0) |
         (a_header.active                   ? 0x40 : 0) |
         (a_header.noactive                 ? 0x80 : 0)
         ;
    ++p;

    *p = (a_header.ignore_self_noactive_calls ? 0x01 : 0) |
         (a_header.directed_call              ? 0x02 : 0) |
         (a_header.from_active_call           ? 0x04 : 0) |
         (a_header.lock_type << 3)
         ;
    ++p;

    if (a_header.id.length() != 18)
      throw std::runtime_error("Incorrect message ID size");
    std::memcpy(p, &a_header.id[0], 18);
    p+=18;

    {
      const size_t l = std::min(a_header.destenition_server.length(), (size_t)32);
      std::memcpy(p, &a_header.destenition_server[0], l);
      std::memset(&p[l], (char)0, 32-l);
      p+=32;
    }

    {
      const size_t l = std::min(a_header.destination_handler.length(), (size_t)16);
      std::memcpy(p, &a_header.destination_handler[0], l);
      std::memset(&p[l], (char)0, 16-l);
      p+=16;
    }

    *(short*)p = fcf::net::le_to_cpu((short)a_header.destination_pair);
    p+=2;
    *(unsigned short*)p = fcf::net::le_to_cpu((unsigned short)a_header.lock_counter);

  }


  event_server_t::event_server_t(server_t& a_server, int a_port)
  : _stop(false), _ready(false), _port(a_port), _server(a_server), _lock_messages(false), _work_guard(boost::asio::make_work_guard(_iocontext))  {
    _locker.reset(event_server::create_locker(_iocontext, *this, _server.get_config().mainServer.c_str()));
  }

  event_server_t::~event_server_t() {
    try {
      stop();
    } catch (...){
    }
  }

  void event_server_t::run(){
    std::unique_lock<std::mutex> lock(_startstop_mutex);

    _thread = std::thread([this](){
        this->_iocontext.run();
    });

    std::vector<server_connection_t>  servers;
    server_connection_t               self_server_connection;
    if (_server.get_config().serverName != _server.get_config().mainServer && !_server.get_config().mainServer.empty()){

      servers = _server.get_config().servers;
      std::vector<server_connection_t>::iterator it;
      it = std::find_if(servers.begin(),
                        servers.end(),
                        [&main_server_name = _server.get_config().mainServer]
                          (server_connection_t a_connection_info) {
                            return a_connection_info.name == main_server_name;
                          }
                        );
      if (it == servers.end())
        throw std::runtime_error("Can't find a self externall event server connection info in config files");

      std::shared_ptr<reconnection_state_t> reconnection_info(new reconnection_state_t());
      reconnection_info->connection_info = *it;
      reconnection_info->post_get_port_callback = [this](reconnection_state_t* a_ri, int a_port, std::string a_address) -> boost::asio::awaitable<void>{
        if (!a_ri->is_first_call_callback)
          co_return;
        a_ri->is_first_call_callback = false;

        server_connection_t elci;
        elci.name     = _server.get_config().serverName;
        elci.address  = a_address;
        elci.port     = a_port;
        try {
          co_await this->_run_external_listener_async(elci);
        } catch(const std::exception& e){
          std::unique_lock<std::mutex> lock(a_ri->mutex);
          a_ri->startup_error = e.what();
          a_ri->condition.notify_all();
        }
        co_return;
      };
      boost::asio::co_spawn(_iocontext, [this, reconnection_info]() -> boost::asio::awaitable<void>{
        co_await _reconnect_to_server(reconnection_info, true);
        co_return;
      }, boost::asio::detached);

      {
        std::unique_lock<std::mutex> lock(reconnection_info->mutex);
        while(!_stop && !reconnection_info->is_first_connect_successful && reconnection_info->startup_error.empty()) {
          reconnection_info->condition.wait_for(lock, std::chrono::milliseconds(100));
        }
        if (!reconnection_info->startup_error.empty()) {
          throw std::runtime_error(reconnection_info->startup_error);
        }
      }

      if (_stop)
        throw std::runtime_error("FAST STOP");

      {
        std::shared_lock<std::shared_mutex> lock(_servers_info_mutex);
        servers = _servers_info;
      }

      std::vector<server_connection_t>::const_iterator self_it =
        std::find_if(servers.begin(),
                     servers.end(),
                     [self_server=_server.get_config().serverName](const server_connection_t& a_val){
                       return self_server == a_val.name;
                     }
                   );
      if (self_it == _server.get_config().servers.end())
        throw std::runtime_error("Can't find the server connection info in the main server configuration");

      self_server_connection = *self_it;
    } else {
      servers = _server.get_config().servers;
      std::vector<server_connection_t>::iterator it = std::find_if(
                        servers.begin(),
                        servers.end(),
                        [&main_server_name = _server.get_config().mainServer]
                          (server_connection_t a_connection_info) {
                            return a_connection_info.name == main_server_name;
                          }
                        );
      if (it == servers.end())
        throw std::runtime_error("Can't find a self externall event server connection info in config files");
      self_server_connection = *it;

      {
        std::unique_lock<std::shared_mutex> lock(_servers_info_mutex);
        _servers_info = servers;
      }

      if (servers.size() > 1) {
        _run_external_listener(self_server_connection);
      }
    }

    for(size_t i = 0; i < servers.size(); ++i) {
      if (servers[i].name == _server.get_config().serverName)
        continue;
      if (servers[i].name == _server.get_config().mainServer)
        continue;
      std::shared_ptr<reconnection_state_t> reconnection_info(new reconnection_state_t());
      boost::asio::co_spawn(_iocontext, [=, this, reconnection_info = reconnection_info, server_connection_info = servers[i]]() -> boost::asio::awaitable<void> {
        reconnection_info->connection_info = server_connection_info;
        co_await _reconnect_to_server(reconnection_info, false);
        co_return;
      },
      boost::asio::detached);

      {
        std::unique_lock<std::mutex> lock(reconnection_info->mutex);
        while(!_stop && reconnection_info->is_during_first_connection_attempt) {
          reconnection_info->condition.wait_for(lock, std::chrono::milliseconds(100));
        }
      }

    }

    {
      std::shared_ptr<running_state_t> running_state(new running_state_t());
      boost::asio::co_spawn(_iocontext, [this, running_state, self_server_connection]() -> boost::asio::awaitable<void>{
        try {
          co_await this->_listener(running_state);
        } catch(std::exception& e){
          running_state->error = e.what();
        }
        running_state->ready = true;
        co_return;
      }, boost::asio::detached);
      while(!running_state->ready){
        std::this_thread::sleep_for(std::chrono::milliseconds(10));
      }
      if (!running_state->error.empty()) {
        throw std::runtime_error(running_state->error);
      }
    }

    {
      bool first = true;
      std::unordered_map<process_identification_t, client_t::client_sp> clients;
      while(true){
        if (first) {
          std::shared_lock<std::shared_mutex> lock(_clients_mutex);
          first = false;
          for(auto& [key, spclient] : _clients){
            if (spclient->process_identification.type == PT_OUTCOMING_COMMANDS && spclient->is_ready_state())
              clients[key] = spclient;
          }
        } else {
          std::shared_lock<std::shared_mutex> lock(_clients_mutex);
          std::vector<process_identification_t> rm;
          for(auto& [key, spclient] : clients) {
            std::unordered_map<process_identification_t, client_t::client_sp>::iterator it =
              _clients.find(key);
            if (it == _clients.end() || it->second.get() != spclient.get()) {
              rm.push_back(key);
            }
            std::unordered_map<process_identification_t, client_t::client_sp>::iterator it_incomming =
              _clients.find({PT_INCOMING_COMMANDS, key.handler, key.process_index, key.server});
            if (it_incomming != _clients.end() && it_incomming->second->is_ready_state()) {
              rm.push_back(key);
            }
          }
          for(process_identification_t& key : rm){
            clients.erase(key);
          }
        }

        if (clients.empty())
          break;

        std::this_thread::sleep_for(std::chrono::milliseconds(10));
      }
    }
  }

  void event_server_t::set_ready_server_state(){
    _ready = true;
    std::unique_lock<std::shared_mutex> lock(_clients_mutex);
    for(auto& [a_pi, a_client] : _clients){
      if (a_pi.type != PT_OUTCOMING_COMMANDS)
        continue;
      _send_ready_state(a_client);
    }
  }

  void event_server_t::stop(){
    std::unique_lock<std::mutex> lock(_startstop_mutex);
    if (_stop)
      return;
    {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      _stop = true;
      std::shared_lock lock_clients(_clients_mutex);
      for(auto& [key, spclient] : _clients){
        std::unique_lock<std::recursive_mutex> lock_client(spclient->mutex);
        spclient->condition.notify_all();
      }
    }
    _iocontext.stop();

    try {
      if (_thread.joinable())
        _thread.join();
    } catch (...){
    }
  }

  void event_server_t::send_message(std::string a_message_name,
                                    std::string& a_message,
                                    std::string a_handler,
                                    int a_process,
                                    int a_timeout_sec){
    client_t::client_sp spclient;
    {
      std::shared_lock<std::shared_mutex> lock(_clients_mutex);
      std::unordered_map<process_identification_t, client_t::client_sp>::iterator client_it = _clients.find(process_identification_t({PT_INNER, a_handler, a_process, _server.get_config().serverName}));
      if (client_it == _clients.end() || !client_it->second->is_ready_state())
        throw std::runtime_error("Can't find or not ready a client for the process (handler: " + a_handler + ", process: " + std::to_string(a_process) + ")");
      spclient = client_it->second;
    }

    message_t::message_sp spmessage(new message_t(spclient));
    spmessage->header.body_size = a_message.length();
    spmessage->header.encryption = false;
    for(size_t i = 0; i < 18; ++i){
      char c = rand() % 128;
      if (!c) c = 1;
      spmessage->header.id.append(1, c);
    }
    spmessage->header.response            = false;
    spmessage->header.error               = false;
    spmessage->header.result              = false;
    spmessage->header.destination_inner   = false;
    spmessage->header.destination_handler = a_handler;
    spmessage->header.destination_pair    = a_process/2;
    spmessage->header.sync_modify         = true;
    spmessage->header.outgoing_message    = true;
    spmessage->processes.push_back(process_identification_t({PT_INNER, a_handler, a_process, _server.get_config().serverName}));
    spmessage->progress.insert(process_identification_t({PT_INNER, a_handler, a_process, _server.get_config().serverName}));
    spmessage->spclient = spclient;

    spmessage->message.resize(a_message.length() + message_header_t::raw_header_size);
    message_header_t::to_raw(&spmessage->message[0], spmessage->header);
    std::memcpy(&spmessage->message[message_header_t::raw_header_size], &a_message[0], a_message.length());

    {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      _messages.insert(std::pair<std::string, message_t::message_sp>(spmessage->header.id, spmessage));
    }

    boost::asio::co_spawn(_iocontext, [this, message = spmessage, client = spclient]() mutable { return this->_send_message(client, message); }, boost::asio::detached);
    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();

    try {
      std::unique_lock<std::recursive_mutex> lock(spmessage->spclient->mutex);
      while(!_stop){
        if (spmessage->header.response)
          break;
        int duration = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count();
        if (duration >= a_timeout_sec) {
          if (a_timeout_sec) {
            throw std::runtime_error(std::string() + "Timeout for server message '" + a_message_name + "'");
          } else {
            break;
          }
        }
        spmessage->spclient->condition.wait_for(lock, std::chrono::seconds(a_timeout_sec - duration));
      }
    } catch (const std::exception& e){
      {
        std::unique_lock<std::recursive_mutex> lock(_mutex);
        _messages.erase(spmessage->header.id);
      }
      throw std::runtime_error(e.what());
    }

    if (_stop)
      return;

    if (spmessage->header.error) {
      std::string error_str;
      try {
        boost::iostreams::stream< boost::iostreams::basic_array_source<char> > sbody(&spmessage->error[0], spmessage->error.size());
        boost::property_tree::ptree tree;
        boost::property_tree::read_json(sbody, tree);
        error_str = tree.get<std::string>("message");
      } catch(...) {
        error_str = "Event error";
      }
      throw std::runtime_error(error_str);
    }

    a_message.assign(&spmessage->message[message_header_t::raw_header_size], spmessage->message.size() - message_header_t::raw_header_size);
  }


  void event_server_t::send_message(boost::property_tree::ptree& a_message_tree,
                                    std::string a_handler,
                                    int a_process,
                                    int a_timeout_sec){
    std::stringstream ss;
    boost::property_tree::write_json(ss, a_message_tree, false);

    client_t::client_sp spclient;
    {
      std::shared_lock<std::shared_mutex> lock(_clients_mutex);
      std::unordered_map<process_identification_t, client_t::client_sp>::iterator client_it = _clients.find(process_identification_t({PT_INNER, a_handler, a_process, _server.get_config().serverName}));
      if (client_it == _clients.end() || !client_it->second->is_ready_state())
        throw std::runtime_error("Can't find or not ready a client for the process (handler: " + a_handler + ", process: " + std::to_string(a_process) + ")");
      spclient = client_it->second;
    }

    message_t::message_sp spmessage(new message_t(spclient));
    spmessage->header.body_size = ss.str().length();
    spmessage->header.encryption = false;
    for(size_t i = 0; i < 18; ++i){
      char c = rand() % 128;
      if (!c) c = 1;
      spmessage->header.id.append(1, c);
    }
    spmessage->header.destenition_server  = _server.get_config().serverName;
    spmessage->header.response            = false;
    spmessage->header.error               = false;
    spmessage->header.result              = false;
    spmessage->header.destination_inner   = false;
    spmessage->header.destination_handler = a_handler;
    spmessage->header.destination_pair    = a_process/2;
    spmessage->header.sync_modify         = true;
    spmessage->header.outgoing_message    = true;
    spmessage->processes.push_back(process_identification_t({PT_INNER, a_handler, a_process, _server.get_config().serverName}));
    spmessage->progress.insert(process_identification_t({PT_INNER, a_handler, a_process, _server.get_config().serverName}));
    spmessage->spclient = spclient;

    spmessage->message.resize(ss.str().length() + message_header_t::raw_header_size);
    message_header_t::to_raw(&spmessage->message[0], spmessage->header);
    std::memcpy(&spmessage->message[message_header_t::raw_header_size], &ss.str()[0], ss.str().length());

    {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      _messages.insert(std::pair<std::string, message_t::message_sp>(spmessage->header.id, spmessage));
    }

    boost::asio::co_spawn(_iocontext, [this, message = spmessage, client = spclient]() mutable { return this->_send_message(client, message); }, boost::asio::detached);
    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();

    try {
      std::unique_lock<std::recursive_mutex> lock(spmessage->spclient->mutex);
      while(!_stop){
        if (spmessage->header.response)
          break;
        int duration = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count();
        if (duration >= a_timeout_sec) {
          if (a_timeout_sec) {
            throw std::runtime_error("Timeout for server message '" + a_message_tree.get<std::string>("name") + "'");
          } else {
            break;
          }
        }
        spmessage->spclient->condition.wait_for(lock, std::chrono::seconds(a_timeout_sec - duration));
      }
    } catch(const std::exception& e) {
      {
        std::unique_lock<std::recursive_mutex> lock(_mutex);
        _messages.erase(spmessage->header.id);
      }
      throw std::runtime_error(e.what());
    }

    if (_stop)
      return;

    if (spmessage->header.error) {
      std::string error_str;
      try {
        boost::iostreams::stream< boost::iostreams::basic_array_source<char> > sbody(&spmessage->error[0], spmessage->error.size());
        boost::property_tree::ptree tree;
        boost::property_tree::read_json(sbody, tree);
        error_str = tree.get<std::string>("message");
      } catch(...) {
        error_str = "Event error";
      }
      throw std::runtime_error(error_str);
    }

    boost::iostreams::stream< boost::iostreams::basic_array_source<char> > sbody(&spmessage->message[message_header_t::raw_header_size], spmessage->message.size() - message_header_t::raw_header_size);
    a_message_tree.clear();
    try {
      boost::property_tree::read_json(sbody, a_message_tree);
    } catch(const std::exception& e){
      throw std::runtime_error(std::string("Incorrect format inner message: ") + e.what());
    }
  }

  void event_server_t::lock_messages(){
    _messages_lock_queue_mutex.lock();
    _lock_messages = true;
  }

  void event_server_t::unlock_messages(){
    _lock_messages = false;
  }

  void event_server_t::process_locked_messages(){
    std::unique_lock<std::recursive_mutex> lock(_messages_lock_mutex);
    try {
      for(auto& [id, message_info] : _locked_messages){
        ::boost::asio::co_spawn(_iocontext,
                              [this, message = message_info.message, client = message_info.client]() mutable -> boost::asio::awaitable<void> {
                                try {
                                  co_await this->_process_message(client, message, client->process_identification.type);
                                } catch(const std::exception& e){
                                  fcf::err << "Event reader error: " << e.what() << " ["
                                           << (client->process_identification.type == PT_INNER             ? "inner connection" :
                                               client->process_identification.type == PT_INCOMING_COMMANDS ? "incomming connection " :
                                                                                                             "outcoming connection " ) << "; "
                                           << " IP: " << client->socket.remote_endpoint().address().to_string() <<  "]"
                                          ;
                                  if (!client->is_ready_state()){
                                    client->close();
                                  }
                                }
                                co_return;
                              },
                              boost::asio::detached);
      }
    } catch(...){
    }
    _locked_messages.clear();
    _messages_lock_queue_mutex.unlock();
  }

  void event_server_t::print_active_messages_of_waiting(){
    std::unique_lock<std::recursive_mutex> lock(_messages_lock_mutex);
    std::set<std::string> names;
    for(auto& [id, message_info] : _locked_messages){
      size_t pos = message_header_t::raw_header_size;
      size_t end = message_info.message->message.size();
      while(pos < end){
        if (message_info.message->message[pos])
          break;
        size_t size = fcf::net::le_to_cpu(*(unsigned int*)(&message_info.message->message[pos+1]));
        pos += 6;
        if (pos + size > end)
          break;
        pos += size;
      }

      boost::iostreams::stream< boost::iostreams::basic_array_source<char> > message_body(&message_info.message->message[pos], message_info.message->message.size() - pos);
      boost::property_tree::ptree tree;
      std::string                 event_name("UNKNOWN");
      try {
        boost::property_tree::read_json(message_body, tree);
        event_name = tree.get<std::string>("name");
      } catch(const std::exception& e) {
      }

      names.insert(event_name);
    }

    std::string str_names;
    bool first = true;
    for(const std::string& event_name : names){
      if (!first)
        str_names += ", ";
      str_names += event_name;
      first = false;
    }

    fcf::err << "Active messages pending processing: " << str_names;
  }

  void event_server_t::_message_to_tree(boost::property_tree::ptree& a_dst, message_t::message_sp a_message){
    size_t pos = message_header_t::raw_header_size;
    size_t end = a_message->message.size();
    while(pos < end){
      if (a_message->message[pos])
        break;
      size_t size = fcf::net::le_to_cpu(*(unsigned int*)(&a_message->message[pos+1]));
      pos += 6;
      if (pos + size > end)
        break;
      pos += size;
    }
    boost::iostreams::stream< boost::iostreams::basic_array_source<char> > message_body(&a_message->message[pos], a_message->message.size() - pos);
    boost::property_tree::read_json(message_body, a_dst);
  }

  void event_server_t::_send_ready_state(client_t::client_sp a_client){
    boost::asio::co_spawn(_iocontext, [this, a_client]() mutable -> boost::asio::awaitable<void> {
      while(!a_client->close_state && !_stop){
        if (!a_client->is_ready_state() || !_ready) {
          co_await fcf::co_sleep(_iocontext, boost::posix_time::milliseconds(10));
          continue;
        }

        process_identification_t pi;
        pi.type = PT_INCOMING_COMMANDS;
        pi.handler = a_client->process_identification.handler;
        pi.process_index = a_client->process_identification.process_index;
        pi.server = a_client->process_identification.server;
        {
          std::unique_lock<std::shared_mutex> lock(_clients_mutex);
          std::unordered_map<process_identification_t, client_t::client_sp>::iterator it = _clients.find(pi);
          if (it == _clients.end() || !it->second->is_ready_state()) {
            lock.unlock();
            co_await fcf::co_sleep(_iocontext, boost::posix_time::milliseconds(10));
            continue;
          }
        }

        {
          std::unique_lock<std::recursive_mutex> lock(a_client->mutex);
          if (a_client->ready_event_sent)
            co_return;
          a_client->ready_event_sent = true;
        }

        try {
          boost::property_tree::ptree tree;
          tree.put("name", "initiaize_fcfserver_ready");
          co_await _send_outgoing_inner_message(a_client, tree);
        } catch(std::exception& e) {
          fcf::err << "Failed send initiaize_fcfserver_ready event: " << e.what();
        }

        break;
      }
      co_return;
    }, boost::asio::detached);
  }

  void event_server_t::_print_connection_error(reconnection_state_t* a_reconnection_state, std::string a_message){
    std::unique_lock<std::mutex> lock(a_reconnection_state->mutex);
    std::chrono::time_point now = std::chrono::system_clock::now();
    if (
        (
          std::chrono::duration_cast<std::chrono::seconds>(now - a_reconnection_state->last_error_time).count() > 60 ||
          a_reconnection_state->connect_iteration == 0
        ) &&
        a_reconnection_state->startup_error.empty()
       )
    {
      a_reconnection_state->last_error_time = now;
      fcf::err << "Error establishing a connection to the '" << a_reconnection_state->connection_info.name
               << "' server for outgoing message [" << a_reconnection_state->connection_info.address
               << " : " << a_reconnection_state->connection_info.port << "]: "
               << a_message;
      fcf::log << "Waiting for connection to '" << a_reconnection_state->connection_info.name
               << "' server [" << a_reconnection_state->connection_info.address
               << " : " << a_reconnection_state->connection_info.port << "] ...";
    }
  }

  void event_server_t::_run_external_listener(const server_connection_t& a_server_connection){
    std::shared_ptr<running_state_t> running_state(new running_state_t());
    boost::asio::co_spawn(_iocontext, [this, running_state, a_server_connection]() -> boost::asio::awaitable<void>{
      try {
        co_await this->_external_listener(a_server_connection, running_state);
      } catch(std::exception& e){
        running_state->error = e.what();
      }
      running_state->ready = true;
      co_return;
    }, boost::asio::detached);
    while(!running_state->ready){
      std::this_thread::sleep_for(std::chrono::milliseconds(10));
    }
    if (!running_state->error.empty()) {
      throw std::runtime_error(running_state->error);
    }
  }

  boost::asio::awaitable<void> event_server_t::_run_external_listener_async(const server_connection_t& a_server_connection){
    std::shared_ptr<running_state_t> running_state(new running_state_t());
    boost::asio::co_spawn(_iocontext, [this, running_state, a_server_connection]() -> boost::asio::awaitable<void>{
      try {
        co_await this->_external_listener(a_server_connection, running_state);
      } catch(std::exception& e){
        running_state->error = e.what();
      }
      running_state->ready = true;
      co_return;
    }, boost::asio::detached);
    while(!running_state->ready){
      co_await fcf::co_sleep(_iocontext, boost::posix_time::milliseconds(10));
    }
    if (!running_state->error.empty()) {
      throw std::runtime_error(running_state->error);
    }
  }

  void event_server_t::_on_open_client(const process_identification_t& a_process_identification){
    try{
      _locker->activate_connection(a_process_identification);
    } catch(const std::exception e){
    }
  }

  void event_server_t::_on_close_client(const process_identification_t& a_process_identification){
    try{
      _locker->deactivate_connection(a_process_identification);
    } catch(const std::exception e){
    }
  }

  boost::asio::awaitable<void> event_server_t::_listener(std::shared_ptr<running_state_t> a_running_state){
    fcf::log << "Start listening control port " << this->_port << " for the local event server";
    boost::asio::any_io_executor executor = co_await boost::asio::this_coro::executor;
    boost::asio::ip::tcp::acceptor acceptor(executor);
    try {
      boost::asio::ip::tcp::endpoint endpoint(boost::asio::ip::address::from_string("127.0.0.1"), (boost::asio::ip::port_type)this->_port);
      acceptor.open(endpoint.protocol());
      acceptor.set_option(boost::asio::ip::tcp::acceptor::reuse_address(true));
      acceptor.bind(endpoint);
      acceptor.listen();
    } catch(const std::exception& e){
      throw std::runtime_error(std::string() + "Error listening on local meessage port: " + e.what());
    }

    a_running_state->ready = true;
    try {
      while(!_stop) {
        boost::asio::ip::tcp::socket socket = co_await acceptor.async_accept(boost::asio::use_awaitable);
        client_t::client_sp client(new client_t(std::move(socket), _iocontext, this));
        boost::asio::co_spawn(executor, [this, client] { return this->_reader(client, PT_INNER); }, boost::asio::detached);
      }
    } catch(const std::exception& e){
      fcf::err << "Server control accept error: " << e.what();
      _server.stop();
    }
    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_external_listener(server_connection_t a_connection_info, std::shared_ptr<running_state_t> a_running_state){
    const std::string server_name  = _server.get_config().serverName;

    if (a_connection_info.port <= 0 || a_connection_info.port > 65535){
      throw std::runtime_error("Incorrect port is in a connection info for this server in the \"servers\" configuration parameter");
    }

    boost::asio::any_io_executor executor = co_await boost::asio::this_coro::executor;
    boost::asio::ip::address address;
    try {
      address = boost::asio::ip::address::from_string(a_connection_info.address);
    } catch(const std::exception& e){
      throw std::runtime_error(std::string() + "The ip address is invalid. the configuration parameter 'servers' is incorrect");
    }

    boost::asio::ip::tcp::acceptor acceptor(executor);
    try {
      boost::asio::ip::tcp::endpoint endpoint(address, (boost::asio::ip::port_type)a_connection_info.port);
      acceptor.open(endpoint.protocol());
      acceptor.set_option(boost::asio::ip::tcp::acceptor::reuse_address(true));
      acceptor.bind(endpoint);
      acceptor.listen();
    } catch(const std::exception& e){
      throw std::runtime_error(std::string() + "Incomming port listening error: " + e.what());
    }

    a_running_state->ready = true;

    try {
      while(!_stop) {
        boost::asio::ip::tcp::socket socket = co_await acceptor.async_accept(boost::asio::use_awaitable);
        client_t::client_sp client(new client_t(std::move(socket), _iocontext, this));
        boost::asio::co_spawn(executor, [this, client] () { return this->_reader(client, PT_INCOMING_COMMANDS); }, boost::asio::detached);
        boost::asio::co_spawn(executor, [this, client] () -> boost::asio::awaitable<void>{
          co_await fcf::co_sleep(_iocontext, boost::posix_time::seconds(10));
          if (!client->is_ready_state() && !client->close_state){
            fcf::err << "Incomming connection system messages has not had time to connect in 10 seconds ";
            client->close();
          }
          co_return;
        }, boost::asio::detached);
      }
    } catch(const std::exception& e){
      fcf::err << "Server control accept error: " << e.what();
      _server.stop();
    }

    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_reader(client_t::client_sp a_client,
                                                       process_type_e a_process_type,
                                                       bool a_reconnect,
                                                       std::shared_ptr<reconnection_state_t> a_reconnection_state,
                                                       bool a_is_main_server_connection){
    char                    buffer[1024*64];
    std::vector<char>       preffix;
    size_t                  beg_buffer = 0;
    bool                    ignore_read = false;
    message_t::message_sp   message(new message_t(a_client));
    size_t                  size = 0;

    unsigned long long      duration = 0;


    while(!_stop) {
      try {
        if (!ignore_read){
          try {
            size = co_await a_client->socket.async_read_some(boost::asio::buffer(buffer), boost::asio::use_awaitable);
          } catch(...){
            break;
          }
        }
        boost::posix_time::ptime t1 = boost::posix_time::microsec_clock::local_time();

        ignore_read = false;

        bool message_format_error = false;
        if (message->header.id.empty() && preffix.size() >= message_header_t::raw_header_size) {
          try {
            message_header_t::from_raw(message->header, &preffix[0]);
          } catch(std::exception& e) {
            fcf::err << "Incorrect format message (" << e.what() << ") on "
                     << (a_client->process_identification.type == PT_INNER             ? "inner event connection " :
                         a_client->process_identification.type == PT_INCOMING_COMMANDS ? "incomming event connection " :
                                                                                         "outcoming event connection " )
                     << " [ip: '" << a_client->socket.remote_endpoint().address().to_string() << "] ";
            preffix.clear();
            message_format_error = true;
          }
        } else if (message->header.id.empty()) {
          if (preffix.size() && preffix.size() + size >= message_header_t::raw_header_size) {
            char raw_header[message_header_t::raw_header_size];
            std::memcpy(&raw_header[0], &preffix[0], preffix.size());
            std::memcpy(&raw_header[preffix.size()], &buffer[beg_buffer], message_header_t::raw_header_size - preffix.size());
            try {
              message_header_t::from_raw(message->header, raw_header);
            } catch(std::exception& e) {
              fcf::err << "Incorrect format message (" << e.what() << ") on "
                       << (a_client->process_identification.type == PT_INNER             ? "inner event connection " :
                           a_client->process_identification.type == PT_INCOMING_COMMANDS ? "incomming event connection " :
                                                                                           "outcoming event connection " )
                       << " [ip: '" << a_client->socket.remote_endpoint().address().to_string() << "] ";
              preffix.clear();
              message_format_error = true;
            }
          } else if (size >= message_header_t::raw_header_size) {
            try {
              message_header_t::from_raw(message->header, &buffer[beg_buffer]);
              if (preffix.capacity() < (message->header.body_size + message_header_t::raw_header_size)){
                preffix.reserve(message->header.body_size + message_header_t::raw_header_size);
              }

            } catch(std::exception& e){
              fcf::err << "Incorrect format message (" << e.what() << ") on "
                       << (a_client->process_identification.type == PT_INNER             ? "inner event connection " :
                           a_client->process_identification.type == PT_INCOMING_COMMANDS ? "incomming event connection " :
                                                                                           "outcoming event connection " )
                       << " [ip: '" << a_client->socket.remote_endpoint().address().to_string() << "] ";
              preffix.clear();
              message_format_error = true;
            }
          }
        }

        if (message_format_error){
          if (a_client->is_ready_state()){
            duration += (boost::posix_time::microsec_clock::local_time() - t1).total_microseconds();
            continue;
          } else {
            throw std::runtime_error("Message format error during connection initialization");
          }
        }

        if (!a_client->is_ready_state() && (message->header.body_size + message_header_t::raw_header_size) > 10000){
          throw std::runtime_error("Message size exceeded during connection initialization");
        }

        if (message->header.body_size + message_header_t::raw_header_size > preffix.size() + size) {
          size_t lsize = preffix.size();
          preffix.resize(lsize + size);
          std::memcpy(&preffix[lsize], &buffer[beg_buffer], size);
          beg_buffer = 0;
          duration += (boost::posix_time::microsec_clock::local_time() - t1).total_microseconds();
          continue;
        }

        std::swap(message->message, preffix);
        size_t lsize = message->message.size();
        message->message.resize(message->header.body_size + message_header_t::raw_header_size);
        size_t suffix_size = (message->header.body_size + message_header_t::raw_header_size) - lsize;
        std::memcpy(&message->message[lsize], &buffer[beg_buffer], suffix_size);

        duration += (boost::posix_time::microsec_clock::local_time() - t1).total_microseconds();
        duration = 0;

        boost::asio::co_spawn(_iocontext,
                              [this, a_client, message, a_process_type]() mutable -> boost::asio::awaitable<void> {
                                try {
                                  co_await this->_process_message(a_client, message, a_process_type);
                                } catch(const std::exception& e){
                                  fcf::err << "Event reader error: " << e.what() << " ["
                                           << (a_client->process_identification.type == PT_INNER             ? "inner connection" :
                                               a_client->process_identification.type == PT_INCOMING_COMMANDS ? "incomming connection " :
                                                                                                               "outcoming connection " ) << "; "
                                           << " IP: " << a_client->socket.remote_endpoint().address().to_string() <<  "]"
                                          ;
                                  if (!a_client->is_ready_state()){
                                    a_client->close();
                                  }
                                }
                                co_return;
                              },
                              boost::asio::detached);

        message.reset(new message_t(a_client));

        if (suffix_size < size){
          beg_buffer += suffix_size;
          size -= suffix_size;
          ignore_read = true;
        } else {
          beg_buffer = 0;
        }
      } catch(std::exception& e) {
        fcf::err << "Event reader error: " << e.what() << " ["
                 << (a_client->process_identification.type == PT_INNER             ? "inner connection" :
                     a_client->process_identification.type == PT_INCOMING_COMMANDS ? "incomming connection " :
                                                                                     "outcoming connection " ) << "; "
                 << " IP: " << a_client->socket.remote_endpoint().address().to_string() <<  "]"
                ;
        if (!a_client->is_ready_state())
          break;
      }
    }

    a_client->close();

    _clear_client(a_client);

    if (a_reconnect){
      if (!_stop) {
        boost::asio::co_spawn(_iocontext, [this, a_reconnection_state, a_is_main_server_connection]() -> boost::asio::awaitable<void> {
          co_await _reconnect_to_server(a_reconnection_state, a_is_main_server_connection);
          co_return;
        }, boost::asio::detached);
      }
    }

    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_connect_to_server(std::shared_ptr<reconnection_state_t> a_reconnection_state, bool a_is_main_server_connection){
    if (a_reconnection_state->connect_iteration == 0){
      fcf::log << "Establishes an outgoing message connection to the '" << a_reconnection_state->connection_info.name << "' server ...";
    }
    ++a_reconnection_state->connect_iteration;

    client_t::client_sp client;
    try {
      boost::asio::ip::address address;
      try {
        address = boost::asio::ip::address::from_string(a_reconnection_state->connection_info.address);
      } catch (const std::exception e) {
        throw std::runtime_error(std::string() + "The ip address is invalid. The configuration parameter 'servers' is incorrect");
      }

      if (a_reconnection_state->connection_info.port <= 0 || a_reconnection_state->connection_info.port > 65535)
        throw std::runtime_error(std::string() + "The port is invalid. The configuration parameter 'servers' is incorrect");
      boost::asio::ip::tcp::endpoint endpoint(address, a_reconnection_state->connection_info.port);
      boost::asio::ip::tcp::socket socket(_iocontext);
      co_await socket.async_connect(endpoint, boost::asio::use_awaitable);
      client.reset(new client_t(std::move(socket), _iocontext, this));
    } catch(const std::exception& e) {
      {
        std::unique_lock<std::mutex> lock(a_reconnection_state->mutex);
        a_reconnection_state->is_during_first_connection_attempt = false;
        a_reconnection_state->condition.notify_all();
      }
      if (!_stop) {
        boost::asio::co_spawn(_iocontext, [this, a_reconnection_state, a_is_main_server_connection]() -> boost::asio::awaitable<void> {
          co_await _reconnect_to_server(a_reconnection_state, a_is_main_server_connection);
          co_return;
        }, boost::asio::detached);
      }
      throw std::runtime_error(e.what());
    }

    boost::asio::co_spawn(_iocontext, [this, client, a_reconnection_state, a_is_main_server_connection] { return this->_reader(client, PT_OUTCOMING_COMMANDS, true, a_reconnection_state, a_is_main_server_connection); }, boost::asio::detached);

    bool error_in_initiaize_connection = false;

    try {
      client->process_identification.type = PT_OUTCOMING_COMMANDS;
      client->process_identification.handler = "";
      client->process_identification.process_index = -1;
      client->process_identification.server = a_reconnection_state->connection_info.name;

      if (a_is_main_server_connection) {
        boost::property_tree::ptree tree;
        tree.put("name", "initiaize_fcfserver_get_port");
        tree.put("server", _server.get_config().serverName);
        co_await _send_outgoing_inner_message(client, tree);

        int port = 0;
        try {
          port = tree.get<int>("port");
        } catch(...){
          throw std::runtime_error("Invalid port value specified when connecting to '" + a_reconnection_state->connection_info.name + "' server");
        }
        if (port <= 0 || port > 65535){
          throw std::runtime_error("Invalid port value specified when connecting to '" + a_reconnection_state->connection_info.name + "' server");
        }

        std::string address;
        try {
          address = tree.get<std::string>("address");
        } catch(...) {
          throw std::runtime_error("Invalid IP address specified when connecting to '" + a_reconnection_state->connection_info.name + "' server");
        }
        if (address.empty()){
          throw std::runtime_error("Invalid IP address specified when connecting to '" + a_reconnection_state->connection_info.name + "' server");
        }

        if (a_reconnection_state->post_get_port_callback){
          co_await a_reconnection_state->post_get_port_callback(a_reconnection_state.get(), port, address);
        }
      }

      try {
        boost::property_tree::ptree tree;
        tree.put("name", "fcfserver_initiaize_connection");
        tree.put("handler", "");
        tree.put("process_index", -1);
        tree.put("server", _server.get_config().serverName);

        {
          std::unique_lock<std::mutex> lock(_tokens_mutex);
          std::string self_token;
          std::map<std::string, std::string>::iterator it_self_token = _self_tokens.find(a_reconnection_state->connection_info.name);
          if (it_self_token == _self_tokens.end()){
            for(size_t i = 0; i < 64; ++i){
              char c = rand() % 128;
              if (!c) c = 1;
              self_token.append(1, c);
            }
            _self_tokens[a_reconnection_state->connection_info.name] = self_token;
          } else {
            self_token = it_self_token->second;
          }
          tree.put("self_token", self_token);

          std::map<std::string, std::string>::iterator it_external_token = _external_tokens.find(a_reconnection_state->connection_info.name);
          if (it_external_token != _external_tokens.end()) {
            tree.put("token", it_external_token->second);
          }
        }

        co_await _send_outgoing_inner_message(client, tree);
      } catch(std::exception& e){
        error_in_initiaize_connection = true;
        a_reconnection_state->connect_iteration = 1;
        throw e.what();
      }

      client->preready = true;

      if (a_is_main_server_connection) {
        //std::vector<server_connection_t>* servers;
        boost::property_tree::ptree tree;
        tree.put("name", "fcfserver_get_environment");
        tree.put("handler", "");
        tree.put("process_index", -1);
        tree.put("server", _server.get_config().serverName);

        co_await _send_outgoing_inner_message(client, tree);

        {
          std::unique_lock<std::shared_mutex> lock(_servers_info_mutex);
          _servers_info.clear();
          for(const auto& ptserver : tree.get_child("servers")) {
            server_connection_t sc;
            sc.name = ptserver.second.get<std::string>("name");
            sc.address = ptserver.second.get<std::string>("address");
            sc.port = ptserver.second.get<int>("port");
            _servers_info.push_back(sc);
          }
        }
      }

      client->set_ready_state();

      _send_ready_state(client);

      fcf::log << "Outgoing message connection to '" << a_reconnection_state->connection_info.name
               << "' server [" << a_reconnection_state->connection_info.address
               << " : " << a_reconnection_state->connection_info.port
               << "] established";

      {
        std::unique_lock<std::mutex> lock(a_reconnection_state->mutex);
        a_reconnection_state->is_first_connect_successful        = true;
        a_reconnection_state->connect_iteration                  = 0;
        a_reconnection_state->last_error_time                    = std::chrono::system_clock::from_time_t(0);
        a_reconnection_state->is_during_first_connection_attempt = false;
        a_reconnection_state->condition.notify_all();
      }

      fcf::log << "Outgoing message connection established with '" << a_reconnection_state->connection_info.name << "' server";

      {
        std::unique_lock<std::shared_mutex> lock(_clients_mutex);
        _clients[client->process_identification] = client;
      }

      {
        std::unique_lock<std::mutex> lock(_tokens_mutex);
        _external_tokens.erase(client->process_identification.server);
      }

    } catch (const std::exception& e){
      if (!error_in_initiaize_connection){
        std::unique_lock<std::mutex> lock(a_reconnection_state->mutex);
        a_reconnection_state->is_during_first_connection_attempt = false;
        a_reconnection_state->condition.notify_all();
      }

      client->close();
    }

    co_return;
  }

  boost::asio::awaitable<void>  event_server_t::_reconnect_to_server(
                                  std::shared_ptr<reconnection_state_t> a_reconnection_state,
                                  bool a_is_main_server_connection) {
    if (a_reconnection_state->connect_iteration != 1){
      co_await fcf::co_sleep(_iocontext, boost::posix_time::milliseconds(1000));
    } else {
      co_await fcf::co_sleep(_iocontext, boost::posix_time::milliseconds(100));
    }

    try{
      co_await _connect_to_server(a_reconnection_state, a_is_main_server_connection);
    } catch(std::exception& e){
      _print_connection_error(a_reconnection_state.get(), e.what());
    }

    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_process_message(client_t::client_sp a_client, message_t::message_sp a_message, process_type_e a_process_type){
    if (!a_client->is_ready_state()){
      a_message->header.encryption = false;
      a_message->header.destenition_server.clear();
      a_message->header.error = false;
      a_message->header.result = false;
      a_message->header.destination_inner = true;
      a_message->header.destination_handler = "";
      a_message->header.destination_pair = -1;
      a_message->header.sync_modify = true;
      a_message->header.enable_connection_error = false;
      a_message->header.active = false;
      a_message->header.noactive = false;
      a_message->header.ignore_self_noactive_calls = false;
      a_message->header.directed_call = false;
      a_message->header.from_active_call = false;
      a_message->header.lock_type = ESTL_NO;
      a_message->header.lock_counter = 0;
    }

    if (a_message->header.response){
      bool complete = false;
      std::map<std::string, message_t::message_sp>::iterator message_it;
      message_t::message_sp spmessage;
      {
        std::unique_lock<std::recursive_mutex> lock(_mutex);
        message_it = _messages.find(a_message->header.id);
        if (message_it == _messages.end()) {
          if (!a_client->is_ready_state()){
            throw std::runtime_error("Event not found: " + a_message->header.id);
          } else {
            fcf::err << "Event not found: " << a_message->header.id << " Process index: " << a_client->process_identification.process_index;
          }
          co_return;
        }
        spmessage = message_it->second;
        spmessage->progress.erase(a_client->process_identification);
      }

      complete = !spmessage->progress.size();
      if (a_message->header.error || spmessage->header.error) {
        if (a_message->header.error) {
          spmessage->header.error = true;
          spmessage->error.assign(a_message->message.begin() + message_header_t::raw_header_size, a_message->message.end());
        }
      } else {
        if (a_message->header.result && !a_message->header.sync_modify) {
          size_t pos = message_header_t::raw_header_size;
          size_t end = a_message->message.size();
          while(pos < end){
            if (a_message->message[pos])
              break;
            size_t size = fcf::net::le_to_cpu(*(unsigned int*)(&a_message->message[pos+1]));
            pos += 6;
            if (pos + size > end)
              break;
            spmessage->result.insert(spmessage->result.end(), a_message->message.begin() + pos - 6, a_message->message.begin() + pos + size);
            pos += size;
          }
        }
        if (spmessage->header.sync_modify) {
          std::swap(spmessage->message, a_message->message);
          spmessage->header.body_size = spmessage->message.size() - message_header_t::raw_header_size;
          message_header_t::to_raw(&spmessage->message[0], spmessage->header);
        }
      }

      if (complete) {
        if (!spmessage->header.outgoing_message) {
          boost::asio::co_spawn(_iocontext, [this, message = spmessage]() mutable { return this->_send_response(message); }, boost::asio::detached);
          if (spmessage->header.sync_modify) {
            std::unique_lock<std::recursive_mutex> lock(spmessage->spclient->mutex);
            spmessage->header.response = true;
            spmessage->spclient->condition.notify_all();
          }
        } else {
          {
            std::unique_lock<std::recursive_mutex> lock(_mutex);
            spmessage->header.response = true;
            _messages.erase(spmessage->header.id);
          }

          std::unique_lock<std::recursive_mutex> lock(spmessage->spclient->mutex);
          spmessage->spclient->condition.notify_all();
        }

      } else if (spmessage->header.sync_modify) {
        std::unique_lock<std::recursive_mutex> lock(spmessage->spclient->mutex);
        spmessage->spclient->condition.notify_all();
      }
    } else {
      if (!a_client->is_ready_state()){
        boost::iostreams::stream< boost::iostreams::basic_array_source<char> > message_body(&a_message->message[message_header_t::raw_header_size], a_message->message.size() - message_header_t::raw_header_size);
        boost::property_tree::ptree tree;
        std::string                 event_name;
        try {
          boost::property_tree::read_json(message_body, tree);
          event_name = tree.get<std::string>("name");
        } catch(const std::exception& e){
          throw std::runtime_error("Incorrect format inner message during connection initialization");
        }

        if (
            (
              !a_client->preready &&
              (
                event_name != "initiaize_fcfserver_get_port" &&
                event_name != "fcfserver_initiaize_connection"
              )
            ) ||
            (
              event_name != "initiaize_fcfserver_get_port" &&
              event_name != "fcfserver_initiaize_connection" &&
              event_name != "fcfserver_get_environment"
            )
          )
        {
          throw std::runtime_error("An invalid message receive during connection initialization");
        }
      }

      if(a_message->header.active && !a_message->header.from_active_call && _lock_messages) {
        std::unique_lock<std::recursive_mutex> lock(_messages_lock_mutex);
        _locked_messages.insert({a_message->header.id, {a_client, a_message}});
        co_return;
      }

      bool local_calls =  (
                            a_message->header.destenition_server.empty() ||
                            a_message->header.destenition_server == _server.get_config().serverName
                          ) &&
                          (!a_message->header.destination_inner);
      bool external_calls = (a_process_type == PT_INNER) &&
                            (
                              a_message->header.destenition_server.empty() ||
                              a_message->header.destenition_server != _server.get_config().serverName
                            ) &&
                            (!a_message->header.destination_inner);

      if (a_process_type == PT_INNER ||
          (a_process_type == PT_INCOMING_COMMANDS && (a_message->header.lock_type == ESTL_SERVER_HANDLER || a_message->header.lock_type == ESTL_SERVER))
         ) {
        if (a_message->header.lock_type != ESTL_NO) {
          std::string handler = a_message->header.lock_type == ESTL_GLOBAL_HANDLER || a_message->header.lock_type == ESTL_SERVER_HANDLER
                                  ? a_message->spclient->process_identification.handler
                                  : std::string();
          event_server::EVENT_SERVER_LOCK_TYPE lock_type = a_message->header.lock_type == ESTL_GLOBAL_HANDLER || a_message->header.lock_type == ESTL_GLOBAL
                                                              ? (_server.get_config().serverName == _server.get_config().mainServer ? event_server::ESLT_INNER_GLOBAL : event_server::ESLT_EXTERNAL)
                                                              : event_server::ESLT_INNER_LOCAL;
          std::string event_name = "UNKNOW";
          try {
            boost::property_tree::ptree tree;
            _message_to_tree(tree, a_message);
            event_name = tree.get<std::string>("name");
            std::string event_key  = tree.find("_fcfLockKey") != tree.not_found() ? tree.get<std::string>("_fcfLockKey") : std::string();
            std::string event_lock_id  = tree.find("_fcfLockId") != tree.not_found() ? tree.get<std::string>("_fcfLockId") : std::string();
            a_message->header.event_name = event_name;
            a_message->header.lock_id = event_lock_id;
            a_message->header.lock_key = event_key;
            co_await _locker->lock( lock_type,
                                    a_message->spclient->process_identification,
                                    handler.c_str(),
                                    event_name.c_str(),
                                    event_lock_id.c_str(),
                                    event_key.c_str(),
                                    a_message->header.lock_counter);
          } catch(const std::exception& e) {
            {
              std::unique_lock<std::recursive_mutex> lock(_mutex);
              _messages.insert(std::pair<std::string, message_t::message_sp>(a_message->header.id, a_message));
            }
            std::string error_str = std::string() + "Invalid lock message [server: " + a_message->spclient->process_identification.server + "; event_name: " + event_name + "]: " + e.what();
            fcf::err << error_str;
            boost::asio::co_spawn(_iocontext, [this, error_str, a_message]() mutable { return this->_send_error(a_message, error_str); }, boost::asio::detached);
            co_return;
          }
        }
      }

      if (local_calls) {
        _server._get_processes_for_message(a_message->processes,
                                           a_message->header.active && !a_message->header.noactive,
                                           a_message->header.active,
                                           a_message->header.destination_pair,
                                           a_message->header.destination_handler,
                                           a_client->process_identification.handler,
                                           a_process_type == PT_INNER && !a_message->header.active && !a_message->header.lock_counter ? a_client->process_identification.process_index : -1);
      }

      {
        std::unique_lock<std::recursive_mutex> lock(_mutex);
        _messages.insert(std::pair<std::string, message_t::message_sp>(a_message->header.id, a_message));
      }

      if (a_message->header.destination_inner) {
        try {
          if (a_message->processes.size())
            fcf::err << "Inner message with process destination!";
          co_await _process_inner_message(a_client, a_message, a_process_type);
        } catch(std::exception& e) {
          fcf::err << "Process inner message error: " << e.what();
          std::string error_str = e.what();
          boost::asio::co_spawn(_iocontext, [this, error_str, a_message]() mutable { return this->_send_error(a_message, error_str); }, boost::asio::detached);
          co_return;
        }
      }

      std::vector<std::pair<process_identification_t, client_t::client_sp> > clients;
      size_t external_calls_size = 0;
      {
        std::shared_lock<std::shared_mutex> lock(_clients_mutex);
        for(process_identification_t& pi : a_message->processes){
          std::unordered_map<process_identification_t, client_t::client_sp>::iterator it = _clients.find(pi);
          clients.push_back({pi, it != _clients.end() ? it->second : client_t::client_sp()});
        }
        if (external_calls){
          for(std::unordered_map<process_identification_t, client_t::client_sp>::iterator it = _clients.begin(); it != _clients.end(); ++it){
            if (it->first.type != PT_OUTCOMING_COMMANDS || !it->second->ready_outcoming)
              continue;
            if ((a_message->header.destenition_server.empty()) ||
                (!a_message->header.destenition_server.empty() && a_message->header.destenition_server == it->first.server)
               ){
              clients.push_back({it->first, it->second});
              ++external_calls_size;
            }
          }
        }
      }

      std::string error_message;
      if (a_message->header.enable_connection_error && external_calls && a_message->header.destenition_server.empty()){
        std::shared_lock<std::shared_mutex> lock(_servers_info_mutex);
        if (_servers_info.size() - 1 != external_calls_size) {
          for(size_t i = 0; i < _servers_info.size(); ++i){
            if (_servers_info[i].name == _server.get_config().serverName)
              continue;
            std::vector<std::pair<process_identification_t, client_t::client_sp> >::const_iterator it =
              std::find_if(clients.begin(), clients.end(), [&server_name = _servers_info[i].name](const std::pair<process_identification_t, client_t::client_sp>& a_item){
                return a_item.first.server == server_name && a_item.first.type == PT_OUTCOMING_COMMANDS;
              });
            if (it == clients.end()){
              error_message += "No connection to '" + _servers_info[i].name + "' server";
              break;
            }
          }
        }
      } else if (a_message->header.enable_connection_error && external_calls && !a_message->header.destenition_server.empty()){
        std::shared_lock<std::shared_mutex> lock(_servers_info_mutex);
        std::vector<std::pair<process_identification_t, client_t::client_sp> >::const_iterator it =
          std::find_if(clients.begin(), clients.end(), [&server_name = a_message->header.destenition_server](const std::pair<process_identification_t, client_t::client_sp>& a_item){
            return a_item.first.server == server_name && a_item.first.type == PT_OUTCOMING_COMMANDS;
          });
        if (it == clients.end()){
          error_message += "No connection to '" + a_message->header.destenition_server + "' server";
        }
      }

      if (!error_message.empty()){
        boost::asio::co_spawn(_iocontext, [this, message = a_message, error_message]() mutable { return this->_send_error(message, error_message); }, boost::asio::detached);
        co_return;
      }

      for(size_t i = 0; i < clients.size(); ++i)
        a_message->progress.insert(clients[i].first);

      if (clients.empty()) {
        message_t::build_responce(*a_message);
        boost::asio::co_spawn(_iocontext, [this, message = a_message]() mutable { return this->_send_response(message); }, boost::asio::detached);
        co_return;
      } else {
        if (!a_message->header.sync_modify) {
          for(size_t client_index = 0; client_index < clients.size(); ++client_index) {
            if (clients[client_index].second) {
              boost::asio::co_spawn(_iocontext, [this, message = a_message, client = clients[client_index].second]() mutable { return this->_send_message(client, message); }, boost::asio::detached);
            } else {
              std::unique_lock<std::recursive_mutex> lock(_mutex);
              a_message->progress.erase(clients[client_index].first);
              if (!a_message->progress.size()){
                boost::asio::co_spawn(_iocontext, [this, a_message]() mutable { return this->_send_response(a_message); }, boost::asio::detached);
              }
            }
          }
        } else {
          boost::asio::co_spawn(_iocontext, [this, message = a_message, clients = std::move(clients)]() mutable { return this->_send_message_sync(std::move(clients), message); }, boost::asio::detached);
        }
      }
    }
    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_send_outgoing_inner_message(
                                  std::string a_server,
                                  boost::property_tree::ptree& a_message_tree) {
    client_t::client_sp spclient;
    {
      std::shared_lock<std::shared_mutex> lock(_clients_mutex);
      std::unordered_map<process_identification_t, client_t::client_sp>::iterator client_it = _clients.find(process_identification_t({PT_OUTCOMING_COMMANDS, "", -1, a_server}));
      if (client_it == _clients.end() || !client_it->second->is_ready_state())
        throw std::runtime_error("Can't find or not ready a client for the process (server: " + a_server + ")");
      spclient = client_it->second;
    }

    co_await _send_outgoing_inner_message(spclient, a_message_tree);

    co_return;
  };


  boost::asio::awaitable<void> event_server_t::_send_outgoing_inner_message(
                                  client_t::client_sp a_client,
                                  boost::property_tree::ptree& a_message_tree){
    std::stringstream ss;
    boost::property_tree::write_json(ss, a_message_tree, false);

    message_t::message_sp spmessage(new message_t(a_client));
    spmessage->header.body_size = ss.str().length();
    spmessage->header.encryption = false;
    for(size_t i = 0; i < 18; ++i){
      char c = rand() % 128;
      if (!c) c = 1;
      spmessage->header.id.append(1, c);
    }
    spmessage->header.destenition_server  = a_client->process_identification.server;
    spmessage->header.response            = false;
    spmessage->header.error               = false;
    spmessage->header.result              = false;
    spmessage->header.destination_inner   = true;
    spmessage->header.destination_handler = a_client->process_identification.handler;
    spmessage->header.destination_pair    = -1;
    spmessage->header.sync_modify         = true;
    spmessage->header.outgoing_message    = true;
    spmessage->processes.push_back(process_identification_t({
      a_client->process_identification.type,
      a_client->process_identification.handler,
      a_client->process_identification.process_index,
      a_client->process_identification.server}));
    spmessage->progress.insert(process_identification_t({
      a_client->process_identification.type,
      a_client->process_identification.handler,
      a_client->process_identification.process_index,
      a_client->process_identification.server}));
    spmessage->spclient = a_client;


    spmessage->message.resize(ss.str().length() + message_header_t::raw_header_size);
    message_header_t::to_raw(&spmessage->message[0], spmessage->header);
    std::memcpy(&spmessage->message[message_header_t::raw_header_size], &ss.str()[0], ss.str().length());

    {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      _messages.insert(std::pair<std::string, message_t::message_sp>(spmessage->header.id, spmessage));
    }
    boost::asio::co_spawn(_iocontext, [this, message = spmessage, client = a_client]() mutable { return this->_send_message(client, message); }, boost::asio::detached);

    std::unique_lock<std::recursive_mutex> lock(spmessage->spclient->mutex);
    while(true){
      {
        std::unique_lock<std::recursive_mutex> lock(_mutex);
        if (!spmessage->progress.count(a_client->process_identification))
          break;
      }
      co_await a_client->condition.co_wait(lock);
    }

    if (_stop)
      co_return;

    if (spmessage->header.error) {
      std::string error_str;
      try {
        boost::iostreams::stream< boost::iostreams::basic_array_source<char> > sbody(&spmessage->error[0], spmessage->error.size());
        boost::property_tree::ptree tree;
        boost::property_tree::read_json(sbody, tree);
        error_str = tree.get<std::string>("message");
      } catch(...) {
        error_str = "Event error";
      }
      throw std::runtime_error(error_str);
    }

    boost::iostreams::stream< boost::iostreams::basic_array_source<char> > sbody(&spmessage->message[message_header_t::raw_header_size], spmessage->message.size() - message_header_t::raw_header_size);
    a_message_tree.clear();
    try {
      boost::property_tree::read_json(sbody, a_message_tree);
    } catch(const std::exception& e){
      throw std::runtime_error(std::string("Incorrect format inner message: ") + e.what());
    }

    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_send_message_sync(std::vector<std::pair<process_identification_t, client_t::client_sp> > a_clients, message_t::message_sp a_message){
    for(size_t client_index = 0; client_index < a_clients.size(); ++client_index){
      client_t::client_sp& client = a_clients[client_index].second;
      if (client){
        co_await this->_send_message(client, a_message);
        {
          std::unique_lock<std::recursive_mutex> lock(a_message->spclient->mutex);
          while(true){
            {
              std::unique_lock<std::recursive_mutex> lock(_mutex);
              if (!a_message->progress.count(client->process_identification))
                break;
            }
            co_await a_message->spclient->condition.co_wait(lock);
          }
        }
      } else {
        std::unique_lock<std::recursive_mutex> lock(_mutex);
        a_message->progress.erase(a_clients[client_index].first);
        if (!a_message->progress.size()) {
          boost::asio::co_spawn(_iocontext, [this, a_message]() mutable { return this->_send_response(a_message); }, boost::asio::detached);
        }
      }
    }
    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_send_message(client_t::client_sp a_client, message_t::message_sp a_message){
    try{
      if (a_message->complete){
        co_return;
      }
      while(a_client->write_state) {
        try { co_await a_client->write_steady_timer.async_wait(boost::asio::use_awaitable); } catch (...){ }
      }
      a_client->write_state = true;
      std::vector<char> buffer(a_message->message);
      co_await boost::asio::async_write(a_client->socket, boost::asio::buffer(buffer),  boost::asio::use_awaitable);
      a_client->write_state = false;
      a_client->write_steady_timer.cancel_one();
    } catch(const std::exception& e) {
      a_client->write_state = false;
      a_client->write_steady_timer.cancel_one();
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      a_message->progress.erase(a_client->process_identification);
      if (!a_message->header.outgoing_message && a_message->header.sync_modify){
        a_message->header.response = true;
        std::unique_lock<std::recursive_mutex> lock(a_message->spclient->mutex);
        a_message->spclient->condition.notify_all();
      } else if (!a_message->progress.size()) {
        if (!a_message->header.outgoing_message){
          boost::asio::co_spawn(_iocontext, [this, a_message]() mutable { return this->_send_response(a_message); }, boost::asio::detached);
        } else {
          a_message->header.response = true;
          auto it = _messages.find(a_message->header.id);
          if (it != _messages.end())
            _messages.erase(it);
          std::unique_lock<std::recursive_mutex> lock(a_message->spclient->mutex);
          a_message->spclient->condition.notify_all();
        }
      }

      a_client->close();
    };
    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_unlock_message(message_t::message_sp a_message){
    if (a_message->spclient->process_identification.type == PT_INNER ||
        (a_message->spclient->process_identification.type == PT_INCOMING_COMMANDS && (a_message->header.lock_type == ESTL_SERVER_HANDLER || a_message->header.lock_type == ESTL_SERVER))
       ) {
      if (a_message->header.lock_type != ESTL_NO) {
        std::string handler = a_message->header.lock_type == ESTL_GLOBAL_HANDLER || a_message->header.lock_type == ESTL_SERVER_HANDLER
                                ? a_message->spclient->process_identification.handler
                                : std::string();
        event_server::EVENT_SERVER_LOCK_TYPE lock_type = a_message->header.lock_type == ESTL_GLOBAL_HANDLER || a_message->header.lock_type == ESTL_GLOBAL
                                                            ? (_server.get_config().serverName == _server.get_config().mainServer ? event_server::ESLT_INNER_GLOBAL : event_server::ESLT_EXTERNAL)
                                                            : event_server::ESLT_INNER_LOCAL;
        std::string event_name = "UNKNOW";
        try {
          std::string event_key;
          std::string event_lock_id;
          if (!a_message->header.event_name.empty()){
            event_name = a_message->header.event_name;
            event_key = a_message->header.lock_key;
            event_lock_id = a_message->header.lock_id;
          } else {
            boost::property_tree::ptree tree;
            _message_to_tree(tree, a_message);
            event_name = tree.get<std::string>("name");
            event_key  = tree.find("_fcfLockKey") != tree.not_found() ? tree.get<std::string>("_fcfLockKey") : std::string();
            event_lock_id  = tree.find("_fcfLockId") != tree.not_found() ? tree.get<std::string>("_fcfLockId") : std::string();
          }
          co_await _locker->unlock( lock_type,
                                  handler.c_str(),
                                  event_name.c_str(),
                                  event_lock_id.c_str(),
                                  event_key.c_str());
        } catch(const std::exception& e) {
          std::string error_str = std::string() + "Invalid unlock message [server: " + a_message->spclient->process_identification.server + "; event_name: " + event_name + "]: " + e.what();
          fcf::err << error_str;
          co_return;
        }
      }
    }
  }

  boost::asio::awaitable<void> event_server_t::_send_response(message_t::message_sp a_message) {

    message_t::message_sp origin_message;
    {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      std::map<std::string, message_t::message_sp>::iterator it = _messages.find(a_message->header.id);
      if (it != _messages.end()){
        origin_message = it->second;
        it->second->complete = true;
        a_message->complete = true;
        _messages.erase(it);
      } else {
        co_return;
      }
    }

    co_await _unlock_message(origin_message);
    origin_message.reset();

    try {
      message_t::build_responce(*a_message);
      while(a_message->spclient->write_state)
        try { co_await a_message->spclient->write_steady_timer.async_wait(boost::asio::use_awaitable); } catch (...) {}
      a_message->spclient->write_state = true;
      co_await boost::asio::async_write(a_message->spclient->socket, boost::asio::buffer(a_message->message),  boost::asio::use_awaitable);
    } catch(...) {
    }
    a_message->spclient->write_state = false;
    a_message->spclient->write_steady_timer.cancel_one();

    _server._release_processes_for_message(a_message->processes, a_message->header.active);
    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_send_error(event_server_t::message_t::message_sp a_message, const std::string a_error_str){

    message_t::message_sp origin_message;

    {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      std::map<std::string, message_t::message_sp>::iterator it = _messages.find(a_message->header.id);
      if (it != _messages.end()){
        origin_message = it->second;
        it->second->complete = true;
        a_message->complete = true;
        _messages.erase(it);
      } else {
        co_return;
      }
    }

    co_await _unlock_message(origin_message);
    origin_message.reset();

    boost::property_tree::ptree tree;
    tree.put("message", a_error_str);
    std::stringstream ss;
    boost::property_tree::write_json(ss, tree, false);
    std::string str = ss.str();

    try {
      a_message->header.error = true;
      a_message->error.assign(str.begin(), str.end());
      message_t::build_responce(*a_message.get());
      while(a_message->spclient->write_state)
        try { co_await a_message->spclient->write_steady_timer.async_wait(boost::asio::use_awaitable); } catch (...) { }
      a_message->spclient->write_state = true;
      co_await boost::asio::async_write(a_message->spclient->socket, boost::asio::buffer(a_message->message),  boost::asio::use_awaitable);
    } catch(...) {
    }
    a_message->spclient->write_state = false;
    a_message->spclient->write_steady_timer.cancel_one();

    _server._release_processes_for_message(a_message->processes, a_message->header.active);
    co_return;
  }

  boost::asio::awaitable<void> event_server_t::_process_inner_message(client_t::client_sp& a_client, message_t::message_sp& a_message, process_type_e a_process_type) {
    boost::iostreams::stream< boost::iostreams::basic_array_source<char> > sbody(&a_message->message[message_header_t::raw_header_size], a_message->message.size() - message_header_t::raw_header_size);
    boost::property_tree::ptree tree;
    try {
      boost::property_tree::read_json(sbody, tree);
    } catch(const std::exception& e){
      throw std::runtime_error(std::string("Incorrect format message: ") + e.what());
    }
    std::string event_name;
    try {
      event_name = tree.get<std::string>("name");
    } catch(const std::exception& e){
      throw std::runtime_error(std::string("Incorrect format message: ") + e.what());
    }

    if (event_name == "fcfserver_restart_processes") {
      std::string handler = tree.find("handler") != tree.not_found() ? tree.get<std::string>("handler") : std::string();
      std::string cause = tree.find("cause") != tree.not_found() ? tree.get<std::string>("cause") : std::string();
      _server._restart_processes(handler, cause);
    } else if (event_name == "initiaize_fcfserver_ready") {
      process_identification_t pi;
      pi.type = PT_OUTCOMING_COMMANDS;
      pi.server = a_client->process_identification.server;
      pi.process_index = -1;
      std::unique_lock<std::shared_mutex> lock(_clients_mutex);
      std::unordered_map<process_identification_t, client_t::client_sp>::iterator it = _clients.find(pi);
      if (it != _clients.end() && it->second->is_ready_state()){
        it->second->ready_outcoming = true;
      } else {
        throw std::runtime_error(std::string() + "A connection ready event has been received but there is no outcoming message connection [" + a_client->socket.remote_endpoint().address().to_string() + "]");
      }
    } else if (event_name == "initiaize_fcfserver_get_port") {
      std::string   server_name       = tree.find("server") != tree.not_found() ? tree.get<std::string>("server") : std::string();
      std::string   address           = a_client->socket.remote_endpoint().address().to_string();
      int           port              = 0;
      bool          close_connection  = false;
      {
        std::shared_lock<std::shared_mutex> lock(_servers_info_mutex);
        std::vector<server_connection_t>::iterator it = std::find_if(
                          _servers_info.begin(),
                          _servers_info.end(),
                          [&server_name, &address]
                            (server_connection_t a_connection_info) {
                              return a_connection_info.name == server_name && a_connection_info.address == address;
                            }
                          );
        if (it == _servers_info.end()) {
          close_connection = true;
        } else {
          port = it->port;
        }
      }

      if (close_connection){
        fcf::wrn << "An initiaize_fcfserver_get_port message was received from a server not declared in the configuration [" << a_client->socket.remote_endpoint().address().to_string() << "]";
        a_client->close();
      }

      tree.put("port", port);
      tree.put("address", address);

      std::stringstream ss;
      boost::property_tree::write_json(ss, tree, false);
      a_message->message.resize(message_header_t::raw_header_size + ss.str().length());
      std::memcpy(&a_message->message[message_header_t::raw_header_size], &ss.str()[0], ss.str().length());
      a_message->header.body_size = ss.str().length();

    } else if (event_name == "fcfserver_initiaize_connection") {
      try {
        bool connect = false;
        bool disconnect = false;
        std::string address = a_client->socket.remote_endpoint().address().to_string();
        a_client->process_identification.type           = a_process_type;
        a_client->process_identification.handler        = tree.get<std::string>("handler");
        a_client->process_identification.process_index  = tree.get<int>("process_index");
        a_client->process_identification.server         = tree.find("server") != tree.not_found() ? tree.get<std::string>("server") : std::string();

        std::string external_token                      = tree.find("self_token") != tree.not_found() ? tree.get<std::string>("self_token") : std::string();
        std::string self_token                          = tree.find("token") != tree.not_found() ? tree.get<std::string>("token") : std::string();

        if  (a_process_type == PT_INCOMING_COMMANDS) {
          std::shared_lock<std::shared_mutex> slock(_servers_info_mutex);
          std::unique_lock<std::mutex> tlock(_tokens_mutex);
          const std::vector<server_connection_t>& servers(_servers_info.size() ? _servers_info : _server.get_config().servers);
          std::vector<server_connection_t>::const_iterator it_servers_info = std::find_if(
                            servers.begin(),
                            servers.end(),
                            [&address]
                              (server_connection_t a_connection_info) {
                                return a_connection_info.address == address;
                              }
                            );
          std::map<std::string, std::string>::iterator it_self_token     = _self_tokens.find(a_client->process_identification.server);
          std::map<std::string, std::string>::iterator it_external_token = _external_tokens.find(a_client->process_identification.server);

          if (it_servers_info == servers.end()) {
            fcf::wrn << "Attempting to connect to the message server from an unregistered IP " << address;
          } else if (it_self_token != _self_tokens.end() && it_self_token->second == self_token &&
                     it_external_token != _external_tokens.end() && it_external_token->second == external_token
                    )
          {
            connect = true;
          } else {
            if (!self_token.empty() && (it_self_token == _self_tokens.end() || it_self_token->second != self_token)){
              //fcf::wrn << "!3";
            } else if (external_token.empty()) {
              //fcf::wrn << "!4";
            } else {
              _external_tokens[a_client->process_identification.server] = external_token;
            }
          }
        } else {
          connect = true;
        }

        if (connect){
          if  (a_process_type == PT_INCOMING_COMMANDS) {
            fcf::log << "Incoming message connection to '" << a_client->process_identification.server
                     << "' server [" << a_client->socket.remote_endpoint().address().to_string()
                     << "] established";
          }
          a_client->set_ready_state();
          std::unique_lock<std::shared_mutex> lock(_clients_mutex);
          _clients[a_client->process_identification] = a_client;
        } else {
          a_client->close();
        }
      } catch(const std::exception& e){
        a_client->close();
        throw std::runtime_error(std::string("Incorrect format message: ") + e.what());
      }
    } else if (event_name == "fcfserver_get_environment") {
      boost::property_tree::ptree ptservers;
      for(const server_connection_t& sc : _server.get_config().servers){
        boost::property_tree::ptree ptitem;
        ptitem.put("name", sc.name);
        ptitem.put("address", sc.address);
        ptitem.put("port", sc.port);
        ptservers.push_back(boost::property_tree::ptree::value_type("", ptitem));
      }
      tree.put_child("servers", ptservers);

      boost::property_tree::ptree pthandlers;
      for(const auto& ep : _server.get_config().endpoints){
        for(const auto& h : ep.handlers){
          pthandlers.push_back(boost::property_tree::ptree::value_type("", h.name));
        }
      }
      tree.put_child("selfHandlers", pthandlers);

      std::stringstream ss;
      boost::property_tree::write_json(ss, tree, false);
      a_message->message.resize(message_header_t::raw_header_size + ss.str().length());
      std::memcpy(&a_message->message[message_header_t::raw_header_size], &ss.str()[0], ss.str().length());
      a_message->header.body_size = ss.str().length();
    } else if (event_name == "fcfserver_lock_event") {
      try {
        process_identification_t pi;
        pi.type = PT_INCOMING_COMMANDS;
        pi.server = tree.get<std::string>("server");
        pi.handler = "";
        pi.process_index = -1;
        co_await _locker->lock(event_server::ESLT_INNER_GLOBAL,
                      pi,
                      tree.get<std::string>("handler").c_str(),
                      tree.get<std::string>("event_name").c_str(),
                      tree.get<std::string>("event_lock_id").c_str(),
                      tree.get<std::string>("key").c_str(),
                      tree.get<unsigned short>("counter"));
      } catch (const std::exception& e) {
        throw std::runtime_error(std::string() + "Invalid process 'fcfserver_lock_event' event: " + e.what());
      }
    } else if (event_name == "fcfserver_unlock_event") {
      try {
        co_await _locker->unlock(event_server::ESLT_INNER_GLOBAL,
                       tree.get<std::string>("handler").c_str(),
                       tree.get<std::string>("event_name").c_str(),
                       tree.get<std::string>("event_lock_id").c_str(),
                       tree.get<std::string>("key").c_str());
      } catch (const std::exception& e) {
        throw std::runtime_error(std::string() + "Invalid process 'fcfserver_unlock_event' event: " + e.what());
      }
    }

    co_return;
 }

  void event_server_t::_clear_client(client_t::client_sp& a_client){
    std::unique_lock<std::recursive_mutex> lock(_mutex);
    {
      std::unique_lock<std::shared_mutex> lock_clients(_clients_mutex);
      auto cit = _clients.find(a_client->process_identification);
      if (cit != _clients.end())
        _clients.erase(cit);
    }

    std::vector<message_t::message_sp> rm;
    for(auto& [key, spmessage] : _messages) {
      std::set<process_identification_t>::iterator curit = spmessage->progress.find(a_client->process_identification);
      if (curit == spmessage->progress.end())
        continue;
      rm.push_back(spmessage);
      if (spmessage->header.enable_connection_error){
        std::string error_message = "{\"message\": \"No connection to '" + a_client->process_identification.server + "' server\"}";
        spmessage->message.assign(error_message.begin(), error_message.end());
        spmessage->header.body_size = error_message.length();
        spmessage->header.error = true;
      }
    }

    for(auto& spmessage : rm) {
      spmessage->progress.erase(a_client->process_identification);
      if (!spmessage->progress.size()) {
        if (!spmessage->header.outgoing_message) {
          boost::asio::co_spawn(_iocontext, [this, spmessage]() mutable { return this->_send_response(spmessage); }, boost::asio::detached);
        } else {
          spmessage->header.response = true;
          spmessage->header.error = true;
          std::string error = "{\"message\": \"Connection broken\"}";
          spmessage->error.assign(error.begin(), error.end());
          _messages.erase(spmessage->header.id);
        }
      }
    }

    std::shared_lock lock_clients(_clients_mutex);
    for(auto& [key, spclient] : _clients){
      std::unique_lock<std::recursive_mutex> lock_client(spclient->mutex);
      spclient->condition.notify_all();
    }
    std::unique_lock<std::recursive_mutex> lock_client(a_client->mutex);
    a_client->condition.notify_all();
  }

}
