#pragma once

#include <atomic>

#include <memory>
#include <list>
#include <map>
#include <set>
#include <unordered_map>
#include <string>
#include <vector>
#include <set>
#include <shared_mutex>
#include <set>
#include <chrono>
#include <boost/asio.hpp>
#include <boost/property_tree/ptree.hpp>
#include "subprocesses.hpp"
#include "configuration.hpp"
#include "event_server/locker.hpp"
#include <fcf/helpers.hpp>
#include <fcf/condition.hpp>

namespace application {
namespace event_server{
  class external_locker_t;
}

  struct running_state_t{
    std::string error;
    bool        ready;
  };

  struct reconnection_state_t{
    server_connection_t                                                                   connection_info;
    std::chrono::time_point<std::chrono::system_clock>                                    last_error_time;
    int                                                                                   connect_iteration;
    bool                                                                                  is_first_connect_successful;
    bool                                                                                  is_during_first_connection_attempt;
    std::string                                                                           startup_error;
    bool                                                                                  is_first_call_callback;
    std::function<
      boost::asio::awaitable<void>
      (reconnection_state_t* a_ri, int a_port, std::string a_address)
      >                                                                                   post_get_port_callback;
    std::condition_variable                                                               condition;
    std::mutex                                                                            mutex;

    reconnection_state_t()
      : last_error_time(std::chrono::system_clock::from_time_t(0))
      , connect_iteration(0)
      , is_first_connect_successful(false)
      , is_during_first_connection_attempt(true)
      , is_first_call_callback(true) {
    }
  };

  class server_t;

  class event_server_t{
    public:
      event_server_t(server_t& a_server, int a_port);
      ~event_server_t();
      void run();
      void set_ready_server_state();
      void stop();
      void send_message(
        boost::property_tree::ptree& a_message_tree,
        std::string a_handler,
        int a_process,
        int a_timeout_sec);
      void send_message(
        std::string a_message_name,
        std::string& a_message,
        std::string a_handler,
        int a_process,
        int a_timeout_sec);
      void lock_messages();
      void unlock_messages();
      void process_locked_messages();

      void print_active_messages_of_waiting();

    private:
      friend event_server::external_locker_t;

      enum lock_type_t{
        ESTL_NO,
        ESTL_GLOBAL = 4,
        ESTL_GLOBAL_HANDLER = 3,
        ESTL_SERVER = 2,
        ESTL_SERVER_HANDLER = 1
      };

      class client_t{
      public:
        typedef std::shared_ptr<client_t>     client_sp;
        typedef std::weak_ptr<client_t>       client_wp;
        boost::asio::ip::tcp::socket          socket;
        process_identification_t              process_identification;
        bool                                  ready_outcoming;
        bool                                  ready_event_sent;
        bool                                  preready;
        bool                                  write_state;
        std::recursive_mutex                  send_mutex;
        std::recursive_mutex                  mutex;
        fcf::condition_t                      condition;
        boost::asio::steady_timer             write_steady_timer;
        bool                                  close_state;
        event_server_t*                       event_server;

        client_t(
          boost::asio::ip::tcp::socket&& a_socket,
          boost::asio::io_context& a_iocontext,
          event_server_t* a_event_server)
          : socket(std::move(a_socket)),
            ready_outcoming(false),
            ready_event_sent(false),
            preready(false),
            write_state(false),
            condition(a_iocontext),
            write_steady_timer(a_iocontext),
            close_state(false),
            event_server(a_event_server),
            _ready(false) {
          }

        void set_ready_state(){
          _ready = true;
          event_server->_on_open_client(process_identification);
        }

        bool is_ready_state(){
          return _ready;
        }

        void close() {
          if (!close_state){
            try {
              close_state = true;
              boost::system::error_code ec;
              socket.shutdown(boost::asio::ip::tcp::socket::shutdown_both, ec);
              socket.close(ec);
              event_server->_on_close_client(process_identification);
            } catch(...) {
            }
          }
        }

      private:
        bool _ready;
      };

      struct message_header_t {
        unsigned int                    body_size;            // 4 bytes.
        bool                            encryption;           // 1 byte.   No Impl
        std::string                     id;                   // 18 bytes.
        std::string                     destenition_server;   // 32 bytes.
        bool                            response;
        bool                            error;
        bool                            result;
        bool                            destination_inner;
        std::string                     destination_handler;
        int                             destination_pair;     // 2 bytes
        bool                            sync_modify;
        bool                            enable_connection_error;
        bool                            active;
        bool                            noactive;
        bool                            ignore_self_noactive_calls;
        bool                            directed_call;
        bool                            from_active_call;
        bool                            outgoing_message;
        lock_type_t                     lock_type;
        unsigned short                  lock_counter;
        std::string                     event_name;
        std::string                     lock_id;
        std::string                     lock_key;


        const static size_t raw_header_size = 76;
        const static int PAIR_PROCESS_NULL = 10000;

        message_header_t()
          : encryption(false)
          , response(false)
          , error(false)
          , result(false)
          , destination_inner(false)
          , destination_pair(0)
          , sync_modify(false)
          , enable_connection_error(false)
          , active(false)
          , noactive(false)
          , ignore_self_noactive_calls(false)
          , directed_call(true)
          , from_active_call(false)
          , outgoing_message(false)
          , lock_type(ESTL_NO)
          , lock_counter(0)
        {
        }

        static void from_raw(message_header_t& a_header, const char* a_raw);
        static void to_raw(char* a_raw, const message_header_t& a_header);
      };

      struct message_t{
        typedef std::shared_ptr<message_t>          message_sp;

        std::vector<process_identification_t>       processes;
        std::set<process_identification_t>          progress;
        message_header_t                            header;
        std::vector<char>                           message;
        std::vector<char>                           error;
        std::vector<char>                           result;
        client_t::client_sp                         spclient;
        bool                                        complete;

        message_t(client_t::client_sp& a_spclient)
          : spclient(a_spclient), complete(false) { }

        static void build_responce(message_t& a_message);
      };

      struct message_process_info_t {
        client_t::client_sp   client;
        message_t::message_sp message;
      };

      void                          _message_to_tree(boost::property_tree::ptree& a_dst, message_t::message_sp a_message);
      void                          _send_ready_state(client_t::client_sp a_client);
      void                          _print_connection_error(reconnection_state_t* a_reconnection_state, std::string a_message);
      void                          _run_external_listener(const server_connection_t& a_server_connection);
      boost::asio::awaitable<void>  _run_external_listener_async(const server_connection_t& a_server_connection);
      boost::asio::awaitable<void>  _connect_to_server(
                                      std::shared_ptr<reconnection_state_t> a_reconnection_state,
                                      bool a_is_main_server_connection);
      boost::asio::awaitable<void>  _reconnect_to_server(
                                      std::shared_ptr<reconnection_state_t> a_reconnection_state,
                                      bool a_is_main_server_connection);
      boost::asio::awaitable<void>  _process_message(
                                      client_t::client_sp a_client,
                                      message_t::message_sp a_message,
                                      process_type_e a_process_type);
      boost::asio::awaitable<void>  _process_inner_message(
                                      client_t::client_sp& a_client,
                                      message_t::message_sp& a_message,
                                      process_type_e a_process_type);
      void                         _clear_client(client_t::client_sp& a_client);
      boost::asio::awaitable<void> _send_outgoing_inner_message(
                                      std::string a_server,
                                      boost::property_tree::ptree& a_message_tree);
      boost::asio::awaitable<void> _send_outgoing_inner_message(
                                      client_t::client_sp a_client,
                                      boost::property_tree::ptree& a_message_tree);
      boost::asio::awaitable<void> _send_message_sync(
                                      std::vector<std::pair<process_identification_t, client_t::client_sp> > a_clients,
                                      message_t::message_sp a_message);
      boost::asio::awaitable<void> _send_message(
                                      client_t::client_sp a_client,
                                      message_t::message_sp a_message);
      boost::asio::awaitable<void> _unlock_message(message_t::message_sp a_message);
      boost::asio::awaitable<void> _send_response(message_t::message_sp a_message);
      boost::asio::awaitable<void> _send_error(
                                      message_t::message_sp a_message,
                                      const std::string a_error_str);
      void                          _on_open_client(const process_identification_t& a_process_identification);
      void                          _on_close_client(const process_identification_t& a_process_identification);
      boost::asio::awaitable<void> _listener(std::shared_ptr<running_state_t> a_running_state);
      boost::asio::awaitable<void> _external_listener(server_connection_t a_connection_info, std::shared_ptr<running_state_t> a_running_state);
      boost::asio::awaitable<void> _reader(client_t::client_sp                  a_client,
                                           process_type_e                       a_process_type,
                                           bool                                 a_reconnect = false,
                                           std::shared_ptr<reconnection_state_t> a_reconnection_state = {},
                                           bool                                 a_is_main_server_connection = false);

      bool                                                                      _stop;
      bool                                                                      _ready;
      std::mutex                                                                _startstop_mutex;
      int                                                                       _port;
      std::thread                                                               _thread;
      server_t&                                                                 _server;
      boost::asio::io_context                                                   _iocontext;
      std::recursive_mutex                                                      _mutex;
      std::shared_mutex                                                         _clients_mutex;
      std::unordered_map<process_identification_t, client_t::client_sp>         _clients;
      std::map<std::string, message_t::message_sp>                              _messages;
      std::recursive_mutex                                                      _messages_lock_mutex;
      std::mutex                                                                _messages_lock_queue_mutex;
      std::vector<server_connection_t>                                          _servers_info;
      std::shared_mutex                                                         _servers_info_mutex;
      bool                                                                      _lock_messages;
      std::map<std::string, message_process_info_t>                             _locked_messages;
      boost::asio::executor_work_guard<boost::asio::io_context::executor_type>  _work_guard;
      std::mutex                                                                _tokens_mutex;
      std::map<std::string, std::string>                                        _self_tokens;
      std::map<std::string, std::string>                                        _external_tokens;
      std::unique_ptr<event_server::locker_i>                                   _locker;
  };

}
