#include <string>
#include <map>
#include <mutex>
#include "../fcf/condition.hpp"
#include "locker.hpp"
#include "../event_server.hpp"
#include "../subprocesses.hpp"


namespace application {
namespace event_server {

  locker_i::~locker_i(){
  };

  class inner_locker_t{
    struct event_info_t{
      process_identification_t  connection_identification;
      std::string               event_lock_id;
      unsigned int              counter;
    };

  public:
    inner_locker_t(boost::asio::io_context& a_iocontext)
      : _iocontext(a_iocontext)
      , _condition(a_iocontext) {
    }

    void activate_connection(const process_identification_t& a_connection_identification){
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      _connections.insert(a_connection_identification);
      _condition.notify_all();
    };

    void deactivate_connection(const process_identification_t& a_connection_identification){
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      _connections.erase(a_connection_identification);

      struct rm_event_t{
        std::string handler;
        std::string event_name;
        std::string event_lock_id;
        std::string key;
      };
      std::vector<rm_event_t> rm_events;
      for(auto& [handler, events] : _locks) {
        for(auto& [event_name, event_info_map] : events) {
          for(auto& [key, event_info] : event_info_map) {
            if (event_info.connection_identification == a_connection_identification){
              rm_events.push_back({handler, event_name, event_info.event_lock_id, key});
            }
          }
        }
      }
      for(const rm_event_t& rm : rm_events) {
        try {
          unlock(rm.handler.c_str(), rm.event_name.c_str(), rm.event_lock_id.c_str(), rm.key.c_str());
        } catch(...){
        }
      }

      _condition.notify_all();
    };

    boost::asio::awaitable<void> lock(const process_identification_t& a_connection_identification, const char* a_handler, const char* a_event_name, const char* a_event_lock_id, const char* a_key, unsigned short a_counter) {
      if (!a_counter)
        a_counter = 1;

      std::unique_lock<std::recursive_mutex> lock(_mutex);
      while(true) {
        if (_connections.find(a_connection_identification) == _connections.end()){
          throw std::runtime_error(std::string() + "Failed to lock '" + a_event_name + "' event because there is no connection to the '" + a_connection_identification.server  + "' server");
        }
        std::map<std::string, std::map<std::string, std::map<std::string, event_info_t> > >::iterator it_handler = _locks.find(a_handler);
        std::map<std::string, std::map<std::string, event_info_t> >::iterator                         it_events;
        std::map<std::string, event_info_t>::iterator                                                 it_event_info;
        if (it_handler == _locks.end()) {
          it_handler = _locks.insert(std::pair<std::string, std::map<std::string, std::map<std::string, event_info_t> > >(a_handler, {})).first;
          auto it_events = it_handler->second.insert(std::pair<std::string, std::map<std::string, event_info_t> >(a_event_name, {})).first;
          it_events->second.insert(std::pair<std::string, event_info_t>(a_key, {a_connection_identification, a_event_lock_id, a_counter}));
        }

        it_events = it_handler->second.find(a_event_name);
        if (it_events == it_handler->second.end()){
          it_events = it_handler->second.insert(std::pair<std::string, std::map<std::string, event_info_t> >(a_event_name, {})).first;
          it_events->second.insert(std::pair<std::string, event_info_t>(a_key, {a_connection_identification, a_event_lock_id, a_counter}));
        }

        it_event_info = it_events->second.find(a_key);
        if (it_event_info == it_events->second.end()){
          it_event_info = it_events->second.insert(std::pair<std::string, event_info_t>(a_key, {a_connection_identification, a_event_lock_id, a_counter})).first;
        }

        if (it_event_info->second.event_lock_id == a_event_lock_id) {
          co_return;
        }

        co_await _condition.co_wait(lock);
      }
    }

    boost::asio::awaitable<void> unlock(const char* a_handler, const char* a_event_name, const char* a_event_lock_id, const char* a_key) {
      std::unique_lock<std::recursive_mutex> lock(_mutex);
      std::map<std::string, std::map<std::string, std::map<std::string, event_info_t> > >::iterator it_handler = _locks.find(a_handler);
      std::map<std::string, std::map<std::string, event_info_t> >::iterator                         it_events;
      std::map<std::string, event_info_t>::iterator                                                 it_event_info;
      if (it_handler == _locks.end()) {
        co_return;
      }

      it_events = it_handler->second.find(a_event_name);
      if (it_events == it_handler->second.end()){
        co_return;
      }

      it_event_info = it_events->second.find(a_key);
      if (it_event_info == it_events->second.end()){
        co_return;
      }

      if (it_event_info->second.event_lock_id != a_event_lock_id) {
        co_return;
      }

      --it_event_info->second.counter;

      if (!it_event_info->second.counter){
        it_events->second.erase(it_event_info);
      }

      if (!it_events->second.size()){
        it_handler->second.erase(it_events);
      }

      if (!it_handler->second.size()){
        _locks.erase(it_handler);
      }

      _condition.notify_all();

      co_return;
    }

  private:
    boost::asio::io_context&                                                            _iocontext;
    std::recursive_mutex                                                                _mutex;
    fcf::condition_t                                                                    _condition;
    std::map<std::string, std::map<std::string, std::map<std::string, event_info_t> > > _locks;
    std::set<process_identification_t>                                                   _connections;
  };


  class external_locker_t {
  public:

    external_locker_t(event_server_t& a_event_server, std::string a_main_server_name)
      : _event_server(a_event_server)
      , _main_server_name(a_main_server_name) {
    }

    boost::asio::awaitable<void> lock(const process_identification_t& a_connection_identification, const char* a_handler, const char* a_event_name, const char* a_event_lock_id, const char* a_key, unsigned short a_counter) {
      boost::property_tree::ptree tree;
      tree.put("name", "fcfserver_lock_event");
      tree.put("server", a_connection_identification.server);
      tree.put("handler", a_handler);
      tree.put("event_name", a_event_name);
      tree.put("event_lock_id", a_event_lock_id);
      tree.put("counter", a_counter);
      tree.put("key", a_key);
      co_await _event_server._send_outgoing_inner_message(_main_server_name, tree);
      co_return;
    }

    boost::asio::awaitable<void> unlock(const char* a_handler, const char* a_event_name, const char* a_event_lock_id, const char* a_key) {
      boost::property_tree::ptree tree;
      tree.put("name", "fcfserver_unlock_event");
      tree.put("handler", a_handler);
      tree.put("event_name", a_event_name);
      tree.put("event_lock_id", a_event_lock_id);
      tree.put("key", a_key);
      co_await _event_server._send_outgoing_inner_message(_main_server_name, tree);
      co_return;
    }

  private:
    event_server_t&           _event_server;
    std::string               _main_server_name;
  };

  class locker_t : public locker_i {
  public:
    locker_t(boost::asio::io_context& a_iocontext, event_server_t& a_event_server, std::string a_main_server_name)
      : _external_locker(new external_locker_t(a_event_server, a_main_server_name))
      ,_inner_local_locker(new inner_locker_t(a_iocontext))
      ,_inner_global_locker(new inner_locker_t(a_iocontext)) {
    }

    virtual ~locker_t(){
    }

    virtual void activate_connection(const process_identification_t& a_connection_identification){
      _inner_local_locker->activate_connection(a_connection_identification);
      _inner_global_locker->activate_connection(a_connection_identification);
    }

    virtual void deactivate_connection(const process_identification_t& a_connection_identification){
      _inner_local_locker->deactivate_connection(a_connection_identification);
      _inner_global_locker->deactivate_connection(a_connection_identification);
    }

    virtual boost::asio::awaitable<void> lock(EVENT_SERVER_LOCK_TYPE a_lock_type,
                                              const process_identification_t& a_connection_identification,
                                              const char* a_handler,
                                              const char* a_event_name,
                                              const char* a_event_lock_id,
                                              const char* a_key,
                                              unsigned short a_counter){
      switch (a_lock_type) {
        case ESLT_INNER_LOCAL:
          co_await _inner_local_locker->lock( a_connection_identification,
                                              a_handler,
                                              a_event_name,
                                              a_event_lock_id,
                                              a_key,
                                              a_counter);
          break;
        case ESLT_INNER_GLOBAL:
          co_await _inner_global_locker->lock(a_connection_identification,
                                              a_handler,
                                              a_event_name,
                                              a_event_lock_id,
                                              a_key,
                                              a_counter);
          break;
        case ESLT_EXTERNAL:
          co_await _external_locker->lock(    a_connection_identification,
                                              a_handler,
                                              a_event_name,
                                              a_event_lock_id,
                                              a_key,
                                              a_counter);
          break;
      }

      co_return;
    }

    virtual boost::asio::awaitable<void> unlock(EVENT_SERVER_LOCK_TYPE a_lock_type,
                                              const char* a_handler,
                                              const char* a_event_name,
                                              const char* a_event_lock_id,
                                              const char* a_key){
      switch (a_lock_type) {
        case ESLT_INNER_LOCAL:
          co_await _inner_local_locker->unlock(   a_handler,
                                                  a_event_name,
                                                  a_event_lock_id,
                                                  a_key);
          break;
        case ESLT_INNER_GLOBAL:
          co_await _inner_global_locker->unlock(a_handler,
                                                a_event_name,
                                                a_event_lock_id,
                                                a_key);
          break;
        case ESLT_EXTERNAL:
          co_await _external_locker->unlock(    a_handler,
                                                a_event_name,
                                                a_event_lock_id,
                                                a_key);
          break;
      }

      co_return;
    }
  private:
    std::unique_ptr<external_locker_t> _external_locker;
    std::unique_ptr<inner_locker_t>    _inner_local_locker;
    std::unique_ptr<inner_locker_t>    _inner_global_locker;

  };

  locker_i* create_locker(boost::asio::io_context& a_iocontext, event_server_t& a_event_server, const char* a_main_server_name){
    return new locker_t(a_iocontext, a_event_server, a_main_server_name);
  }

}
}
