#pragma once

#include <pwd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef __APPLE__
#include <libproc.h>
#endif

#include <array>
#include <cstring>
#include <string>
#include <regex>
#include <vector>
#include <chrono>
#include <thread>
#include <fstream>
#include <sys/wait.h>
#include "../except.hpp"
#include "../log.hpp"
#include "../process.hpp"

extern char **environ;

namespace fcf {
namespace details {
namespace {

  void wait_handler(pid_t a_pid)
  {
    int status;
    while(check_pid(a_pid)){
      waitpid(a_pid, &status, WNOHANG);
      std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }
  }


  std::string exec(const char* cmd) {
    std::array<char, 265> buffer;
    std::string result;
    std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(cmd, "r"), pclose);
    if (!pipe) {
      throw std::runtime_error((std::string() + "Faild execute command:" + cmd + "popen() failed!").c_str());
    }
    while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr) {
      result += buffer.data();
    }
    return result;
  }



  struct subprocess_params_t {
    subprocess_params_t(const char* a_name, std::vector<std::string> a_argv, std::vector<std::pair<std::string, std::string> > a_envv)
    {
      argv.push_back(std::strcpy(new char[strlen(a_name)+1], a_name));
      for(size_t i = 0; i < a_argv.size(); ++i)
        argv.push_back(std::strcpy(new char[a_argv[i].length() + 1], a_argv[i].c_str()));
      argv.push_back((char*)0);

      for(size_t i = 0; i < a_envv.size(); ++i){
        std::string val = a_envv[i].first + "=" + a_envv[i].second;
        envv.push_back(std::strcpy(new char[val.length() + 1], val.c_str()));
      }
      envv.push_back((char*)0);
    }

    ~subprocess_params_t() {
      for(size_t i = 0; i < argv.size()-1; ++i)
        delete[] argv[i];
      for(size_t i = 0; i < envv.size()-1; ++i)
        delete[] envv[i];
    }

    std::vector<char*> argv;
    std::vector<char*> envv;
  };



  inline pid_t create_process(const char* a_name,
                              std::vector<std::string> a_argv,
                              std::vector<std::pair<std::string, std::string> > a_envv,
                              const char* a_user,
                              const char* a_password) {
    a_name = a_name ? a_name : "";
    a_user = a_user ? a_user : "";
    a_password = a_password ? a_password : "";

    int uid = -1;
    int gid = -1;
    subprocess_params_t params(a_name, a_argv, a_envv);
    if (a_user && a_user[0]) {
      std::string output = exec((std::string() + "/usr/bin/id " + a_user).c_str());
      output = output.substr(0, output.find('\n'));
      std::regex  regex("[ ]*uid=([0-9]*)[^\\)]*\\)[ ]*gid=([0-9]*)(.*)");
      std::smatch match;
      if (!std::regex_match(output, match, regex)){
        throw fcf::except::nouserinfo(a_user);
      }
      uid = std::stoi(match[1]);
      gid = std::stoi(match[2]);
    }

    struct stat stproc;
    std::string cmdline;
    if (a_name[0])
      cmdline += std::string() + "\"" + a_name + "\"";

    for(size_t i = 0; i < a_argv.size(); ++i){
      if (i != 0 || (i == 0 && a_name[0]))
        cmdline += " ";
      cmdline +=  a_argv[i];
    }

    pid_t pid = fork();
    if (!pid) {
      if ((a_user && a_user[0]) && setgid(gid) == -1){
        fcf::err << "setgid(). Failed to change user. Check the rights to change user. ";
        fcf::err << strerror(errno);
        exit(90);
      }
      if ((a_user && a_user[0]) && setuid(uid) == -1){
        fcf::err << "setuid(). Failed to change user. Check the rights to change user. ";
        fcf::err << strerror(errno);
        exit(90);
      }
      execve(a_name, &params.argv[0], &params.envv[0]);
      fcf::err << "Failed execute command";
      fcf::err << strerror(errno);
      exit(1);
    }

    if (pid == -1 || !pid){
      throw fcf::except::process_start(cmdline.c_str());
    }


    std::thread thread(wait_handler, pid);
    thread.detach();

    return pid;
  }

  inline void close_process(pid_t a_pid) {
    if (!a_pid)
      return;
    ::kill(a_pid, SIGKILL);
    ::kill(a_pid, SIGTERM);
  }

  inline bool check_pid(pid_t a_pid) {
    return 0 == kill(a_pid, 0);
  }

  #ifdef __APPLE__
    inline unsigned long long get_process_memory_size(int a_pid){
      proc_taskallinfo pti;
      if (proc_pidinfo(a_pid, PROC_PIDTASKALLINFO, 0, &pti, sizeof(pti)) >= sizeof(pti))
        return (long)pti.ptinfo.pti_resident_size;
      return 0;
    }
  #else
    inline unsigned long long get_process_memory_size(int a_pid) {
      std::string buff;
      std::ifstream fs(std::string("/proc/") + std::to_string(a_pid) + "/stat");
      std::getline(fs, buff);
      size_t cn(0), pos(0), lstpos(0), l(buff.length());
      for (size_t i = 0; i < l; ++i) {
        if (buff[i] == ' ') {
          lstpos = pos;
          pos = i + 1;
          ++cn;
        }
        if (cn == 24)
          return std::stoul(buff.substr(lstpos, i - lstpos)) * getpagesize();
      }
      return 0;
    }
  #endif

}
}
}
