#include <climits>
#include <string>
#include <iostream>
#include <sstream>

#define BOOST_BIND_GLOBAL_PLACEHOLDERS
#include <boost/bind.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <fcf/string.hpp>
#include <fcf/log.hpp>


#include "server.hpp"

namespace application {

  server_t::server_t(const server_configuration_t& a_config, const global_configuration_t& a_global_config)
    : _stop(false), _exit(false), _config(a_config), _global_config(a_global_config) {
    for(int i = 0; i < a_config.endpoints.size(); ++i){
      _endpoints.push_back( endpoint_sptr_t(new endpoint_t(a_config, _global_config, i)) );
    }
    _event_server.reset(new event_server_t(*this, a_config.innerControlPort));
  }

  server_t::~server_t(){
  }

  void server_t::run() {
    std::unique_lock<std::mutex> lock(_start_stop_mutex);
    if (_stop)
      return;

    _event_server->run();

    if (_stop)
      return;

    for(int i = 0; i < _endpoints.size(); ++i){
      if (_stop)
        return;
      _endpoints[i]->run();
    }

    _event_server->set_ready_server_state();
  }

  void server_t::stop() {
    std::unique_lock<std::mutex> lock(_start_stop_mutex);
    if (_stop)
      return;
    _stop = true;
    fcf::log << "Stop server ";
    fcf::log << "--------------------------------------------";

    for(int i = 0; i < _endpoints.size(); ++i)
      _endpoints[i]->stop();

    _event_server->stop();

    _exit = true;
  }

  bool server_t::is_stop(){
    return _stop;
  }

  bool server_t::is_exit(){
    return _exit;
  }

  event_server_t* server_t::get_event_server(){
    return _event_server.get();
  }

  void server_t::_get_processes_for_message(std::vector<process_identification_t>& a_dst,
                                            bool a_only_active_processes,
                                            bool a_active_flag,
                                            int a_destenition_pair,
                                            std::string a_handler,
                                            std::string a_ignore_handler,
                                            int a_ignore_process){
    for(endpoint_sptr_t& endpoint : _endpoints)
      endpoint->get_processes_for_message(a_dst,
                                          a_only_active_processes,
                                          a_active_flag,
                                          a_destenition_pair,
                                          a_handler,
                                          a_ignore_handler,
                                          a_ignore_process);
  }

  void server_t::_release_processes_for_message(const std::vector<process_identification_t>& a_procs, bool a_active){
    for(endpoint_sptr_t& endpoint : _endpoints)
      endpoint->release_processes_for_message(a_procs, a_active);
  }

  void server_t::_restart_processes(std::string a_handler, const std::string& a_cause) {
    if (!a_handler.empty())
      fcf::log << "Restarting processes at the request of the handler for handler " << a_handler << " (" << a_cause << ")";
    else
      fcf::log << "Restarting processes at the request of the handler (" << a_cause << ")";

    if (a_handler.empty()){
      for(endpoint_sptr_t& endpoint : _endpoints){
        endpoint->restart_handlers();
      }
    } else {
      for(endpoint_sptr_t& endpoint : _endpoints){
        endpoint->restart_handler(a_handler);
      }
    }
  }

  const server_configuration_t& server_t::get_config(){
    return _config;
  }

}
