#define BOOST_BIND_GLOBAL_PLACEHOLDERS

#include <set>
#include <chrono>

#include <boost/bind/bind.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <fcf/except.hpp>
#include <fcf/net/socket.hpp>
#include <fcf/log.hpp>
#include <fcf/process.hpp>

#include "server.hpp"
#include "subprocesses.hpp"

extern application::server_sptr_t pserver_g;

namespace {
#ifdef WIN32
#define ENABLE_SUBPROCESS 0
#else
#define ENABLE_SUBPROCESS 1
#endif
}

namespace application {

subprocesses_t::subprocesses_t(const global_configuration_t& a_global_config, size_t a_maxmem, size_t a_maxmem_long_request, std::string a_handler, std::string a_server_name)
  : _global_config(a_global_config), _maxmem(a_maxmem), _maxmem_long_request(a_maxmem_long_request), _stop(false), _handler(a_handler), _proc_counter_sf(0), _proc_counter_nsf(0), _proc_counter(0), _server_name(a_server_name) {

}

subprocesses_t::~subprocesses_t() {
  stop();
}

int subprocesses_t::append(const char* a_name,
                            std::vector<std::string>                          a_argv1,
                            std::vector<std::string>                          a_argv2,
                            std::vector<std::pair<std::string, std::string> > a_envv,
                            std::vector<std::string>                          a_node_options1,
                            std::vector<std::string>                          a_node_options2,
                            const char*                                       a_user,
                            const char*                                       a_password,
                            int                                               a_start_timeout,
                            int                                               a_stop_timeout,
                            bool                                              a_fast_switching,
                            bool                                              a_leading) {
  _processes.push_back({a_name, a_argv1, a_envv, a_node_options1, a_user,a_password, a_start_timeout, a_stop_timeout, (pid_t)0, 0, 0, 0, true, a_fast_switching, a_leading});
  _processes.push_back({a_name, a_argv2, a_envv, a_node_options2, a_user,a_password, a_start_timeout, a_stop_timeout, (pid_t)0, 0, 0, 0, false, a_fast_switching, a_leading});
  return (int)_processes.size() - 1;
}

void subprocesses_t::run(){
  for(int i = 0; i < _processes.size(); ++i) {
    if (i%2 == 0) {
      _restart_process(i, false);
      boost::property_tree::ptree ptree;
      ptree.put("name", "init_long_live_data");
      try {
        pserver_g->get_event_server()->send_message(ptree, _handler, i, 5);
      } catch (const std::exception& e){
        fcf::err << "Can't process init_long_live__data event for the process index: " << i <<  ")";
        throw e;
      }

    }
  }

  _restart_flags.resize(_processes.size() / 2);
  _observers.resize(_processes.size() / 2);
  for(size_t i = 0; i < (_processes.size() / 2); ++i){
    _observers[i] = std::thread(&subprocesses_t::_observer, this, i*2, i*2+1);
  }

}

void subprocesses_t::stop(){
  if (_stop)
    return;
  _stop = true;

  {
    std::unique_lock<std::mutex> lock(_mutex);
    _condition.notify_all();
  }

  for(std::thread& thread : _observers){
    if (thread.joinable())
      thread.join();
  }

  for(int i = 0; i < _processes.size(); ++i){
    _stop_process(i, true);
  }
}

int subprocesses_t::get_process(SUBPROCESS_FAST_SWITCHING a_lp){
  int result = -1;
  int number;
  if (a_lp == SFS_YES) {
    number = _proc_counter_sf.fetch_add(1) % _processes.size();
  } else if (a_lp == SFS_NO) {
    number = _proc_counter_nsf.fetch_add(1) % _processes.size();
  } else {
    number = _proc_counter.fetch_add(1) % _processes.size();
  }

  while(result == -1) {
    for(int i = number; i < _processes.size(); ++i) {
      if (!_processes[i].active || !_processes[i].pid)
        continue;
      if (_processes[i].fast_switching && a_lp == SFS_NO)
        continue;
      if (!_processes[i].fast_switching && a_lp == SFS_YES)
        continue;
      result = i;
      break;
    }
    if (result == -1){
      for(int i = 0; i < number; ++i) {
        if (!_processes[i].active || !_processes[i].pid)
          continue;
        if (_processes[i].fast_switching && a_lp == SFS_NO)
          continue;
        if (!_processes[i].fast_switching && a_lp == SFS_YES)
          continue;
        result = i;
        break;
      }
    }
  }

  _processes[result].counter.fetch_add(1);
  return result;
}

subprocesses_t::subprocesses_state_t& subprocesses_t::get_process_info(int a_index){
  return _processes[a_index];
}

void subprocesses_t::release_process(int a_index){
  if (_processes[a_index].counter.load() <= 0) { // reserve protect
    fcf::err << "Incorrect computation task counter";
    return;
  }
  _processes[a_index].counter.fetch_sub(1);
}

void subprocesses_t::get_processes_for_message(std::vector<process_identification_t>& a_dst,
                                               bool                                   a_only_active_processes,
                                               bool                                   a_active_flag,
                                               int                                    a_destenition_pair,
                                               std::string                            a_handler,
                                               std::string                            a_ignore_handler,
                                               int                                    a_ignore_process){
  if (!a_handler.empty() && a_handler != _handler) {
    return;
  }
  std::unique_lock<std::mutex> lock(_active_messages_mutex, std::defer_lock);
  if (a_active_flag)
    lock.lock();

  for(int i = 0; i < _processes.size(); ++i) {
    if (a_ignore_handler == _handler && i == a_ignore_process)
      continue;
    if (a_destenition_pair != -1 && i/2 != a_destenition_pair)
      continue;
    if (!_processes[i].pid)
      continue;
    if (a_only_active_processes && !_processes[i].active)
      continue;
    _processes[i].low_counter.fetch_add(1);
    if (a_active_flag)
      _processes[i].low_counter_active.fetch_add(1);
    a_dst.push_back(process_identification_t({PT_INNER, _handler, i, _server_name}));
  }
}

void subprocesses_t::release_processes_for_message(const std::vector<process_identification_t>& a_procs, bool a_active){
  std::unique_lock<std::mutex> lock(_active_messages_mutex, std::defer_lock);
  if (a_active)
    lock.lock();

  for(int i = 0; i < a_procs.size(); ++i) {
    if (a_procs[i].handler != _handler){
      continue;
    }
    if (_processes[a_procs[i].process_index].low_counter.load() <= 0) { // reserve protect
      fcf::err << "Incorrect computation task counter for messages";
    } else {
      _processes[a_procs[i].process_index].low_counter.fetch_sub(1);
    }
    if (a_active){
      if (_processes[a_procs[i].process_index].low_counter_active.load() <= 0) { // reserve protect
        fcf::err << "Incorrect computation task counter for messages";
      } else {
        _processes[a_procs[i].process_index].low_counter_active.fetch_sub(1);
      }
    }
  }

  if (a_active)
    _active_messages_condition.notify_all();
}

bool subprocesses_t::is_active(int a_procNum) {
  return _processes[a_procNum].active;
}

void subprocesses_t::restart(){
  std::unique_lock<std::mutex> lock(_mutex);
  for(size_t i = 0; i < _restart_flags.size(); ++i)
    _restart_flags[i] = true;
}

request_processing_information_t subprocesses_t::get_request_processing_information(std::string a_path, std::string a_context) {
  request_processing_information_t result({true, -1, -1, -1, -1});
  boost::property_tree::ptree ptree;
  ptree.put("name", "fcf_get_path_info");
  ptree.put("path", a_path);
  ptree.put("context", a_context);

  int process_index = get_process(SFS_NO_MATTER);

  try {
    pserver_g->get_event_server()->send_message(ptree, _handler, process_index, 5);
  } catch(std::exception& e){
    release_process(process_index);
    return result;
  }

  release_process(process_index);

  result.short_mode = ptree.get<bool>("shortMode");

  return result;
}

void subprocesses_t::_stop_process(int a_index, bool a_hide_errors, bool a_force){ // no exceptions
  if (!_processes[a_index].pid)
    return;

  pid_t pid          = _processes[a_index].pid;
  pid_t subpid       = _processes[a_index].subpid;
  int   stop_timeout = !a_force ? _processes[a_index].stop_timeout : 5;

  std::chrono::system_clock::time_point start = std::chrono::system_clock::now();
  try {
    std::thread thread = std::thread([this, a_index, stop_timeout](){
      this->_send_close(a_index, stop_timeout);
    });
    thread.detach();

    while(true){
      if (!fcf::check_pid(pid)){
        _processes[a_index].pid = 0;
        break;
      }
      int wsec = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count();
      if (wsec > stop_timeout)
        throw std::runtime_error("Process termination timeout exceeded [PID:" + std::to_string(subpid ? subpid : pid) + "]");
      std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }
    _processes[a_index].pid = 0;
    _processes[a_index].subpid = 0;
  } catch(const std::exception& e) {
    if (!a_hide_errors)
      fcf::err << "Invalid process termination: " << e.what();
    fcf::close_process(subpid);
    fcf::close_process(pid);
    _processes[a_index].pid = 0;
    _processes[a_index].subpid = 0;
    return;
  }
}

void subprocesses_t::_restart_process(int a_index, bool a_restart){
  _stop_process(a_index, false);

  std::vector<std::string> args;
  for(size_t i = 0; i < _processes[a_index].node_options.size(); ++i)
    args.push_back(_processes[a_index].node_options[i]);

  bool check_port = false;
  for(size_t i = 0; i < _processes[a_index].argv.size(); ++i){
    if (_processes[a_index].argv[i] == "--port") {
      check_port = true;
      args.push_back(_processes[a_index].argv[i]);
    } else if (check_port) {
      check_port = false;
      if (std::stoi(_processes[a_index].argv[i]) <= 0){
        int port = fcf::net::get_free_tcp_port("NODEJS application data port(dataPorts parameter)", _global_config.free_ports);
        args.push_back(std::to_string(port));
        _processes[a_index].real_port = port;
      } else {
        args.push_back(_processes[a_index].argv[i]);
        _processes[a_index].real_port = std::stoi(_processes[a_index].argv[i]);
      }
    } else {
      args.push_back(_processes[a_index].argv[i]);
    }
  }

  std::string cmdline = _processes[a_index].name + " ";

  for(size_t i = 0; i < args.size(); ++i)
    cmdline += args[i] + " ";

  fcf::log << "Start process: " << cmdline;

  pid_t subpid = 0;
  pid_t pid = fcf::create_process(_processes[a_index].name.c_str(),
                                  args,
                                  _processes[a_index].envv,
                                  _processes[a_index].user.c_str(),
                                  _processes[a_index].password.c_str());
  if (pid) {
    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();
    try {
      while(true){
        if (!fcf::check_pid(pid)){
          throw fcf::except::process_start(cmdline.c_str());
        }

        try{
          pid_t reqpid = _send_require_subpid(a_index, 5);
          if (ENABLE_SUBPROCESS)
            subpid = reqpid;
          if (reqpid)
            break;
        } catch (const std::exception& e){
          std::this_thread::sleep_for(std::chrono::milliseconds(50));
        }

        int stat = 0;
        if (!fcf::check_pid(pid)){
          throw fcf::except::process_start(cmdline.c_str());
        }

        int wsec = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count();
        if (_processes[a_index].start_timeout > 0 && wsec >= _processes[a_index].start_timeout){
          throw fcf::except::process_timeout(cmdline.c_str());
        }
      }
    } catch(const std::runtime_error& e) {
      if (ENABLE_SUBPROCESS)
        fcf::close_process(subpid);
      fcf::close_process(pid);
      throw e;
    } catch(const std::exception& e) {
      if (ENABLE_SUBPROCESS)
        fcf::close_process(subpid);
      fcf::close_process(pid);
      throw e;
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    fcf::log << "Process started successfully (PID:" << pid << "): " << cmdline;
  } else {
    _processes[a_index].pid = 0;
    throw fcf::except::process_start(cmdline.c_str());
  }

  _processes[a_index].pid = pid;
  if (ENABLE_SUBPROCESS)
    _processes[a_index].subpid = subpid;

}

void subprocesses_t::_observer(int a_first_procindex, int a_second_procindex) {
  int   current_procindex = a_first_procindex;
  int   reserve_procindex = a_second_procindex;
  int   is_first_iteration = true;
  bool  restart_error = false;

  struct message_locker_t{
    message_locker_t(::application::event_server_t* a_event_server)
      : _event_server(a_event_server), _lock(true){
      _event_server->lock_messages();
    }

    ~message_locker_t(){
      unlock();
    }

    void unlock(){
      if (!_lock)
        return;
      _lock = false;
      pserver_g->get_event_server()->unlock_messages();
      pserver_g->get_event_server()->process_locked_messages();
    }
    ::application::event_server_t* _event_server;
    bool _lock;
  };


  {
    std::unique_lock<std::mutex> lock(_mutex);
    _condition.wait_for(lock, std::chrono::seconds(5));
    if (_stop)
      return;
  }

  while(true) {
    std::unique_lock<std::mutex> lock(_mutex);

    if (_stop)
      return;

    try {
      bool restart_flag = _restart_flags[a_first_procindex/2];
      lock.unlock();
      size_t memsize = ENABLE_SUBPROCESS ? fcf::get_process_memory_size(_processes[current_procindex].subpid) / 1000000
                                         : fcf::get_process_memory_size(_processes[current_procindex].pid) / 1000000;
      lock.lock();

      if (restart_flag)
        _restart_flags[a_first_procindex/2] = false;

      if (_stop)
        return;

      if (is_first_iteration)
        fcf::log << "Process memory size  (PID:" << (ENABLE_SUBPROCESS ? _processes[current_procindex].subpid : _processes[current_procindex].pid) << ") " << memsize << "MB";

      size_t maxmem = _processes[current_procindex].fast_switching ? _maxmem : _maxmem_long_request;

      bool switch_over = !memsize ||
                         (memsize > maxmem && maxmem) ||
                         !_processes[current_procindex].pid ||
                         restart_flag ||
                         restart_error;

      if (switch_over){
        lock.unlock();
          _restart_process(reserve_procindex, true);

          // wait processing active messages
          while(true) {
            std::unique_lock<std::mutex> lock(_wait_active_messages_mutex);
            message_locker_t message_locker(pserver_g->get_event_server());
            bool             overtime = false;

            std::chrono::time_point<std::chrono::system_clock> start =  std::chrono::system_clock::now();
            while(true){
              std::unique_lock<std::mutex> lock(_active_messages_mutex);

              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count() >  60){
                overtime = true;
                break;
              }

              if (_stop){
                return;
              }

              bool break_loop =  true;
              for(auto& process : _processes) {
                if (process.low_counter_active.load()){
                  break_loop = false;
                  break;
                }
              }

              if (break_loop)
                break;

              _active_messages_condition.wait_for(lock, std::chrono::milliseconds(100));
            }

            if (overtime){
              fcf::err << "Timeout waiting for active events to complete when switching processes";
              pserver_g->get_event_server()->print_active_messages_of_waiting();
              message_locker.unlock();
              std::this_thread::sleep_for(std::chrono::milliseconds(1000));
              continue;
            }

            // send get_long_live_data & accept_long_live_data events
            bool valid_get_long_live_data = true;
            std::string long_live_data_event = "{\"name\": \"get_long_live_data\"}";
            try {
              pserver_g->get_event_server()->send_message("get_long_live_data", long_live_data_event, _handler, current_procindex, 5);
            } catch (const std::exception& e){
              if (!_stop)
                fcf::err << "Can't process get_long_live__data event for the process index: " << current_procindex <<  ")";
              valid_get_long_live_data = false;
            }
            if (valid_get_long_live_data){
              try{
                pserver_g->get_event_server()->send_message("accept_long_live_data", long_live_data_event, _handler, reserve_procindex, 5);
              } catch (const std::exception& e){
                if (!_stop)
                  fcf::err << "Can't process get_long_live__data event for the process index: " << reserve_procindex <<  ")";
              }
            } else {
              boost::property_tree::ptree ptree;
              ptree.put("name", "init_long_live_data");
              try {
                pserver_g->get_event_server()->send_message(ptree, _handler, reserve_procindex, 5);
              } catch (const std::exception& e){
                if (!_stop)
                  fcf::err << "Can't process init_long_live__data event for the process index: " << reserve_procindex <<  ")";
                throw e;
              }
            }

            _processes[reserve_procindex].active = true;
            _processes[current_procindex].active = false;
            break;
          }

          std::chrono::system_clock::time_point starttm = std::chrono::system_clock::now();
          while(true) {
            while(_processes[current_procindex].counter.load() || _processes[current_procindex].low_counter.load()) {
              std::this_thread::sleep_for(std::chrono::milliseconds(10));
              if (_stop)
                return;
              int dsec = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - starttm).count();
              if (restart_flag || _restart_flags[a_first_procindex/2] || dsec > _processes[current_procindex].stop_timeout) {
                _stop_process(current_procindex, false, true);
                break;
              }
            }
            bool ready = true;
            for(size_t i = 0; i < 10; ++i){
              std::this_thread::sleep_for(std::chrono::milliseconds(100));
              if (_processes[current_procindex].counter.load() || _processes[current_procindex].low_counter.load()){
                ready = false;
                break;
              }
            }
            if (ready)
              break;
          }

          std::swap(current_procindex, reserve_procindex);
          _stop_process(reserve_procindex, false);

        lock.lock();
      }

      restart_error = false;
    } catch (std::exception& e) {
      if (!_stop)
        fcf::err << "Observer failed:" << e.what();
      restart_error = true;
    } catch (...) {
      if (!_stop)
        fcf::err << "Unknown Exception in process observer";
      restart_error = true;
    }

    if (!lock.owns_lock())
      lock.lock();

    _condition.wait_for(lock, std::chrono::seconds(1));

    is_first_iteration = false;
  }
}

void subprocesses_t::_send_close(int a_proc_index, int a_timeout){
  boost::property_tree::ptree ptree;
  ptree.put("name", "fcf_exit");
  try {
    pserver_g->get_event_server()->send_message(ptree, _handler, a_proc_index, a_timeout);
  } catch(std::exception& e) {
  }
}


pid_t subprocesses_t::_send_require_subpid(int a_proc_index, int a_timeout){
  boost::property_tree::ptree ptree;
  ptree.put("name", "fcf_check_ready");
  pserver_g->get_event_server()->send_message(ptree, _handler, a_proc_index, a_timeout);
  if (!ptree.get<bool>("ready"))
    throw std::runtime_error("Process not ready");
  return ptree.get<int>("pid");
}

}
