#pragma once
#include <atomic>
#include <vector>
#include <string>
#include <thread>
#include <mutex>
#include <condition_variable>
#include "global_configuration.hpp"

namespace application {

  #ifdef WIN32
  typedef int pid_t;
  #else
  typedef ::pid_t pid_t;
  #endif

  enum SUBPROCESS_FAST_SWITCHING {
    SFS_YES,
    SFS_NO,
    SFS_NO_MATTER
  };

  enum process_type_e {
    PT_INNER,
    PT_INCOMING_COMMANDS,
    PT_OUTCOMING_COMMANDS
  };

  struct process_identification_t{
    process_type_e  type;
    std::string     handler;
    int             process_index;
    std::string     server;

    inline bool operator< (const process_identification_t& a_struct) const {
      return this->type < a_struct.type                   ? true :
             this->type > a_struct.type                   ? false :
             this->server < a_struct.server               ? true :
             this->server > a_struct.server               ? false :
             this->handler < a_struct.handler             ? true :
             this->handler > a_struct.handler             ? false :
             this->process_index < a_struct.process_index ? true :
                                                            false;
    }

    inline bool operator== (const process_identification_t& a_struct) const {
      return this->type == a_struct.type &&
             this->server == a_struct.server &&
             this->handler == a_struct.handler &&
             this->process_index == a_struct.process_index;
    }
  };

} // application namespace

namespace std {
  template <>
  struct hash<application::process_identification_t>{
    size_t operator()(const application::process_identification_t& a_pi) const{
      return hash<std::string>()(a_pi.handler + std::to_string(a_pi.process_index));
    }
  };

} // std namespace


namespace application {

  struct request_processing_information_t {
    bool  short_mode  = true;
    int   client_socket_timeout   = -1;
    int   max_request_timeout     = -1;
    int   server_socket_timeout   = -1;
    int   max_response_timeout    = -1;
  };

  class subprocesses_t {
  public:
    struct subprocesses_state_t {
      std::string                                       name;
      std::vector<std::string>                          argv;
      std::vector<std::pair<std::string, std::string> > envv;
      std::vector<std::string>                          node_options;
      std::string                                       user;
      std::string                                       password;
      int                                               port;
      int                                               real_port;
      int                                               start_timeout;
      int                                               stop_timeout;
      pid_t                                             pid;
      pid_t                                             subpid;
      std::atomic_int32_t                               counter;
      std::atomic_int32_t                               low_counter;
      std::atomic_int32_t                               low_counter_active;
      bool                                              active;
      bool                                              fast_switching;
      bool                                              leading;

      inline subprocesses_state_t()
        : pid(0), subpid(0), counter(0), low_counter(0), low_counter_active(0) { }

      inline subprocesses_state_t(
        const char*                                                                         a_name,
        std::vector<std::string>                                                            a_argv,
        std::vector<std::pair<std::string, std::string> >                                   a_envv,
        std::vector<std::string>                                                            a_node_options,
        const char*                                                                         a_user,
        const char*                                                                         a_password,
        int                                                                                 a_start_timeout,
        int                                                                                 a_stop_timeout,
        pid_t                                                                               a_pid,
        std::atomic_int32_t                                                                 a_counter,
        std::atomic_int32_t                                                                 a_low_counter,
        std::atomic_int32_t                                                                 a_low_counter_active,
        bool                                                                                a_active,
        bool                                                                                a_fast_switching,
        bool                                                                                a_leading)
      : name(a_name), argv(a_argv), envv(a_envv), node_options(a_node_options),
        user(a_user), password(a_password), start_timeout(a_start_timeout),
        stop_timeout(a_stop_timeout), pid(a_pid), counter(a_counter.load()),
        low_counter(a_low_counter.load()), low_counter_active(a_low_counter_active.load()),
        active(a_active), fast_switching(a_fast_switching), leading(a_leading)
      { }

      inline subprocesses_state_t(const subprocesses_state_t& a_struct)
        : name(a_struct.name), argv(a_struct.argv), envv(a_struct.envv), node_options(a_struct.node_options),
          user(a_struct.user), password(a_struct.password), start_timeout(a_struct.start_timeout),
          stop_timeout(a_struct.stop_timeout), pid(a_struct.pid), counter(a_struct.counter.load()),
          low_counter(a_struct.low_counter.load()), low_counter_active(a_struct.low_counter_active.load()),
          active(a_struct.active), fast_switching(a_struct.fast_switching), leading(a_struct.leading)
      { }
    };

    subprocesses_t(const global_configuration_t& a_global_config, size_t a_maxmem, size_t a_maxmem_long_request, std::string a_handler, std::string a_server_name);
    ~subprocesses_t();
    int                     append(const char*                                        a_name,
                                   std::vector<std::string>                           a_argv1,
                                   std::vector<std::string>                           a_argv2,
                                   std::vector<std::pair<std::string, std::string> >  a_envv,
                                   std::vector<std::string>                           a_node_options1,
                                   std::vector<std::string>                           a_node_options2,
                                   const char*                                        a_user,
                                   const char*                                        a_password,
                                   int                                                a_start_timeout,
                                   int                                                a_stop_timeout,
                                   bool                                               a_fast_switching,
                                   bool                                               a_leading);
    void                    run();
    void                    stop();
    int                     get_process(SUBPROCESS_FAST_SWITCHING a_lp = SFS_NO_MATTER);
    subprocesses_state_t&   get_process_info(int a_index);
    void                    release_process(int a_index);
    void                    get_processes_for_message(std::vector<process_identification_t>&  a_dst,
                                                      bool                                    a_only_active_processes,
                                                      bool                                    a_active_flag,
                                                      int                                     a_destenition_pair,
                                                      std::string                             a_handler,
                                                      std::string                             a_ignore_handler,
                                                      int                                     a_ignore_process);
    void                    release_processes_for_message(const std::vector<process_identification_t >& a_procs, bool a_active);
    bool                    is_active(int a_procNum);
    void                    restart();
    request_processing_information_t get_request_processing_information(std::string a_path, std::string a_context);

  private:
    void                    _stop_process(int a_index, bool a_hide_errors, bool a_force = false); // no exceptions
    void                    _restart_process(int a_index, bool a_restart);
    void                    _observer(int a_first_procindex, int a_second_procindex);
    void                    _send_close(int a_port, int a_timeout);
    pid_t                   _send_require_subpid(int a_proc_index, int a_timeout);

    const global_configuration_t        _global_config;
    size_t                              _maxmem;
    size_t                              _maxmem_long_request;
    bool                                _stop;
    std::vector<bool>                   _restart_flags;
    std::vector<std::thread>            _observers;
    std::vector<subprocesses_state_t>   _processes;
    std::mutex                          _mutex;
    std::mutex                          _start_mutex;
    std::mutex                          _active_messages_mutex;
    std::condition_variable             _active_messages_condition;
    std::mutex                          _wait_active_messages_mutex;
    std::condition_variable             _condition;
    std::string                         _handler;
    std::atomic_uint32_t                _proc_counter_sf;
    std::atomic_uint32_t                _proc_counter_nsf;
    std::atomic_uint32_t                _proc_counter;
    std::string                         _server_name;


  };

  typedef std::shared_ptr<subprocesses_t> subprocesses_sptr_t;

}
